from platypush.backend.sensor import SensorBackend


class SensorSerialBackend(SensorBackend):
    """
    This backend listens for new events from sensors connected through a serial
    interface (like Arduino) acting as a wrapper for the ``serial`` plugin.

    Requires:

        * The :mod:`platypush.plugins.serial` plugin configured
    """

    def __init__(self, **kwargs):
        super().__init__(plugin='serial', **kwargs)


# vim:sw=4:ts=4:et:



from platypush.plugins import Plugin, action


class ClipboardPlugin(Plugin):
    """
    Plugin to programmatically copy strings to your system clipboard
    and get the current clipboard content.

    Requires:
        * **pyperclip** (``pip install pyperclip``)
    """

    @action
    def copy(self, text):
        """
        Copies a text to the OS clipboard

        :param text: Text to copy
        :type text: str
        """
        import pyperclip
        pyperclip.copy(text)


    @action
    def paste(self):
        """
        Get the current content of the clipboard
        """
        import pyperclip
        return pyperclip.paste()


# vim:sw=4:ts=4:et:




"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

from platypush.plugins import action
from platypush.plugins.google import GooglePlugin


class GoogleYoutubePlugin(GooglePlugin):
    """
    YouTube plugin
    """

    scopes = ['https://www.googleapis.com/auth/youtube.readonly']

    # See https://developers.google.com/youtube/v3/getting-started#part
    _default_parts = ['snippet']

    # See https://developers.google.com/youtube/v3/getting-started#resources
    _default_types = ['video']

    def __init__(self, *args, **kwargs):
        super().__init__(scopes=self.scopes, *args, **kwargs)

    @action
    def search(self, parts=None, query='', types=None, max_results=25, **kwargs):
        """
        Search for YouTube content.

        :param parts: List of parts to get (default: snippet).
            See the `Getting started - Part <https://developers.google.com/youtube/v3/getting-started#part>`_.
        :type parts: list[str] or str

        :param query: Query string (default: empty string)
        :type query: str

        :param types: List of types to retrieve (default: video).
            See the `Getting started - Resources <https://developers.google.com/youtube/v3/getting-started#resources>`_.
        :type types: list[str] or str

        :param max_results: Maximum number of items that will be returned (default: 25).
        :type max_results: int

        :param kwargs: Any extra arguments that will be transparently passed to the YouTube API.
            See the `Getting started - parameters <https://developers.google.com/youtube/v3/docs/search/list#parameters>`_.

        :return: A list of YouTube resources.
            See the `Getting started - Resource <https://developers.google.com/youtube/v3/docs/search#resource>`_.
        """

        parts = parts or self._default_parts[:]
        if isinstance(parts, list):
            parts = ','.join(parts)

        types = types or self._default_types[:]
        if isinstance(types, list):
            types = ','.join(types)

        service = self.get_service('youtube', 'v3')
        result = service.search().list(part=parts, q=query, type=types,
                                       maxResults=max_results,
                                       **kwargs).execute()

        events = result.get('items', [])
        return events


# vim:sw=4:ts=4:et:



import time

from platypush.plugins import Plugin, action


class MidiPlugin(Plugin):
    """
    Virtual MIDI controller plugin. It allows you to send custom MIDI messages
    to any connected devices.

    Requires:

        * **python-rtmidi** (``pip install python-rtmidi``)
    """

    _played_notes = set()

    def __init__(self, device_name='Platypush virtual MIDI output',
                 *args, **kwargs):
        """
        :param device_name: MIDI virtual device name (default: *Platypush virtual MIDI output*)
        :type device_name: str
        """

        import rtmidi
        super().__init__(*args, **kwargs)

        self.device_name = device_name
        self.midiout = rtmidi.MidiOut()
        available_ports = self.midiout.get_ports()

        if available_ports:
            self.midiout.open_port(0)
            self.logger.info('Initialized MIDI plugin on port 0')
        else:
            self.open_virtual_port(self.device_name)
            self.logger.info('Initialized MIDI plugin on virtual device {}'.
                         format(self.device_name))


    @action
    def send_message(self, values, *args, **kwargs):
        """
        :param values: Values is expected to be a list containing the MIDI command code and the command parameters - see reference at https://ccrma.stanford.edu/~craig/articles/linuxmidi/misc/essenmidi.html
        :type values: list[int]

        Available MIDI commands:
             * ``0x80``    Note Off
             * ``0x90``    Note On
             * ``0xA0``    Aftertouch
             * ``0xB0``    Continuous controller
             * ``0xC0``    Patch change
             * ``0xD0``    Channel Pressure
             * ``0xE0``    Pitch bend
             * ``0xF0``    Start of system exclusive message
             * ``0xF1``    MIDI Time Code Quarter Frame (Sys Common)
             * ``0xF2``    Song Position Pointer (Sys Common)
             * ``0xF3``    Song Select
             * ``0xF6``    Tune Request (Sys Common)
             * ``0xF7``    End of system exclusive message
             * ``0xF8``    Timing Clock (Sys Realtime)
             * ``0xFA``    Start (Sys Realtime)
             * ``0xFB``    Continue (Sys Realtime)
             * ``0xFC``    Stop (Sys Realtime)
             * ``0xFE``    Active Sensing (Sys Realtime)
             * ``0xFF``    System Reset (Sys Realtime)

        :param args: Extra args that will be passed to ``rtmidi.send_message``
        :param kwargs: Extra kwargs that will be passed to ``rtmidi.send_message``
        """

        self.midiout.send_message(values, *args, **kwargs)


    @action
    def play_note(self, note, velocity, duration=0):
        """
        Play a note with selected velocity and duration.

        :param note: MIDI note in range 0-127 with #60 = C4
        :type note: int

        :param velocity: MIDI note velocity in range 0-127
        :type velocity: int

        :param duration: Note duration in seconds. Pass 0 if you don't want the note to get off
        :type duration: float
        """

        self.send_message([0x90, note, velocity])  # Note on
        self._played_notes.add(note)

        if duration:
            time.sleep(duration)
            self.send_message([0x80, note, 0])  # Note off
            self._played_notes.remove(note)


    @action
    def release_note(self, note):
        """
        Release a played note.

        :param note: MIDI note in range 0-127 with #60 = C4
        :type note: int
        """

        self.send_message([0x80, note, 0])  # Note off
        self._played_notes.remove(note)


    @action
    def release_all_notes(self):
        """
        Release all the notes being played.
        """

        played_notes = self._played_notes.copy()
        for note in played_notes:
            self.release_note(note)


    @action
    def query_ports(self):
        """
        :returns: dict: A list of the available MIDI ports with index and name
        """

        import rtmidi
        in_ports = rtmidi.MidiIn().get_ports()
        out_ports = rtmidi.MidiOut().get_ports()

        return {
            'in': { i: port for i, port in enumerate(in_ports) },
            'out': { i: port for i, port in enumerate(out_ports) },
        }


# vim:sw=4:ts=4:et:





from platypush.plugins import Plugin, action


class HomeseerPlugin(Plugin):
    """
    This plugin allows you interact with an existing HomeSeer setup,
    query and control connected devices.

    Requires:

        * **pyhomeseer** (``pip install git+https://github.com/legrego/PyHomeSeer``)
    """

    def __init__(self, host, username=None, password=None, *args, **kwargs):
        """
        :param host: IP or hostname of your HomeSeer hub
        :type host: str

        :param username: HomeSeer username
        :type username: str

        :param password: HomeSeer password
        :type password: str
        """

        super().__init__(*args, **kwargs)

        self.host = host
        self.username = username
        self.password = password
        self._client = None

    def _get_client(self):
        from pyhomeseer.homeseer_client import HomeSeerClient

        if not self._client:
            self._client = HomeSeerClient(host=self.host,
                                          username=self.username,
                                          password=self.password)

        return self._client


    @action
    def query_devices(self, ref=None, location=None):
        """
        Get a list of devices connected to HomeSeer with their status

        :param ref: Device reference. If not set, all the devices will be queried
        :type ref: int

        :param location: Device location. If not set, all the devices will be queried
        :type location: str
        """

        client = self._get_client()

        if ref is not None:
            devices = client.get_devices(ref=ref)
        elif location is not None:
            devices = client.get_devices(location=location)
        else:
            devices = client.get_devices()

        return [
            {
                attr: getattr(dev, attr)
                for attr in ['ref','name','location','value','status']
            }
            for dev in devices
        ]


    @action
    def control(self, ref, value=None, label=None):
        """
        Control a HomeSeer connected device.

        :param ref: Device reference
        :type ref: int

        :param value: If set, then control the device with this specific int value
        :type value: int

        :param label: If set, then control the device with this specific label
            (e.g. 'On' or 'Off')
        :type label: str
        """

        if value is None and label is None:
            raise RuntimeError('Please specify either value or label')

        client = self._get_client()

        if value is not None:
            return client.control(ref=ref, value=value)
        return client.control(ref=ref, label=label)


# vim:sw=4:ts=4:et:



import time

from threading import Thread

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.light import LightStatusChangeEvent


class LightHueBackend(Backend):
    """
    This backend will periodically check for the status of your configured
    Philips Hue light devices and trigger events when the status of a device
    (power, saturation, brightness or hue) changes.

    Triggers:

        * :class:`platypush.message.event.light.LightStatusChangeEvent` when the
            status of a lightbulb changes

    Requires:

        * The :class:`platypush.plugins.light.hue.LightHuePlugin` plugin to be
            active and configured.
    """

    _DEFAULT_POLL_SECONDS = 10

    def __init__(self, poll_seconds=_DEFAULT_POLL_SECONDS, *args, **kwargs):
        """
        :param poll_seconds: How often the backend will poll the Hue plugin for
            status updates. Default: 10 seconds
        :type poll_seconds: float
        """

        super().__init__(*args, **kwargs)
        self.poll_seconds = poll_seconds


    def _get_lights(self):
        plugin = get_plugin('light.hue')
        if not plugin:
            plugin = get_plugin('light.hue', reload=True)

        return plugin.get_lights().output

    def _listener(self):
        def _thread():
            lights = self._get_lights()

            while not self.should_stop():
                try:
                    lights_new = self._get_lights()

                    for light_id, light in lights_new.items():
                        event_args = {}
                        state = light.get('state')
                        prev_state = lights.get(light_id, {}).get('state', {})

                        if 'on' in state and state.get('on') != prev_state.get('on'):
                            event_args['on'] = state.get('on')
                        if 'bri' in state and state.get('bri') != prev_state.get('bri'):
                            event_args['bri'] = state.get('bri')
                        if 'sat' in state and state.get('sat') != prev_state.get('sat'):
                            event_args['sat'] = state.get('sat')
                        if 'hue' in state and state.get('hue') != prev_state.get('hue'):
                            event_args['hue'] = state.get('hue')
                        if 'ct' in state and state.get('ct') != prev_state.get('ct'):
                            event_args['ct'] = state.get('ct')

                        if event_args:
                            event_args['light_id'] = light_id
                            event_args['light_name'] = light.get('name')
                            self.bus.post(LightStatusChangeEvent(**event_args))

                    lights = lights_new
                except Exception as e:
                    self.logger.exception(e)
                finally:
                    time.sleep(self.poll_seconds)

        return _thread

    def run(self):
        super().run()

        while not self.should_stop():
            try:
                poll_thread = Thread(target=self._listener())
                poll_thread.start()
                poll_thread.join()
            except Exception as e:
                self.logger.exception(e)
                time.sleep(self.poll_seconds)

# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.message import Mapping
from platypush.message.response import Response


class GoogleDriveResponse(Response):
    pass


class GoogleDriveFile(Mapping):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 type: str,
                 id: str,
                 name: str,
                 mime_type: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(id=id, name=name, type=type,
                         mime_type=mime_type, *args, **kwargs)


# vim:sw=4:ts=4:et:



import datetime
import logging
import inspect
import json
import time

logger = logging.getLogger(__name__)


class Message(object):
    """ Message generic class """

    class Encoder(json.JSONEncoder):
        @staticmethod
        def parse_numpy(obj):
            try:
                import numpy as np
            except ImportError:
                return

            if isinstance(obj, np.floating):
                return float(obj)
            if isinstance(obj, np.integer):
                return int(obj)
            if isinstance(obj, np.ndarray):
                return obj.tolist()

            return

        @staticmethod
        def parse_datetime(obj):
            if isinstance(obj, datetime.datetime) or \
                    isinstance(obj, datetime.date) or \
                    isinstance(obj, datetime.time):
                return obj.isoformat()

        def default(self, obj):
            value = self.parse_datetime(obj)
            if value is not None:
                return value

            if isinstance(obj, set):
                return list(obj)

            value = self.parse_numpy(obj)
            if value is not None:
                return value

            return super().default(obj)

    def __init__(self, timestamp=None, *args, **kwargs):
        self.timestamp = timestamp or time.time()

    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """

        return json.dumps({
            attr: getattr(self, attr)
            for attr in self.__dir__()
            if (attr != '_timestamp' or not attr.startswith('_'))
            and not inspect.ismethod(getattr(self, attr))
        }, cls=self.Encoder).replace('\n', ' ')

    def __bytes__(self):
        """
        Overrides the bytes() operator, converts the message into
        its JSON-serialized UTF-8-encoded representation
        """
        return str(self).encode('utf-8')

    @classmethod
    def parse(cls, msg):
        """
        Parse a generic message into a key-value dictionary
        Params:
            msg -- Original message - can be a dictionary, a Message,
                   or a string/bytearray, as long as it's valid UTF-8 JSON
        """

        if isinstance(msg, cls):
            msg = str(msg)
        if isinstance(msg, bytes) or isinstance(msg, bytearray):
            msg = msg.decode('utf-8')
        if isinstance(msg, str):
            # noinspection PyBroadException
            try:
                msg = json.loads(msg.strip())
            except:
                logger.warning('Invalid JSON message: {}'.format(msg))

        assert isinstance(msg, dict)

        if '_timestamp' not in msg:
            msg['_timestamp'] = time.time()

        return msg

    @classmethod
    def build(cls, msg):
        """
        Builds a Message object from a dictionary.
        Params:
            msg -- The message as a key-value dictionary, Message object or JSON string
        """
        from platypush.utils import get_message_class_by_type

        msg = cls.parse(msg)
        msgtype = get_message_class_by_type(msg['type'])
        if msgtype != cls:
            return msgtype.build(msg)


class Mapping(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return k in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def pop(self, *args):
        return self.__dict__.pop(*args)

    def __cmp__(self, dict_):
        return self.__cmp__(dict_)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __str__(self):
        return str(self.__dict__)


# vim:sw=4:ts=4:et:



from typing import Optional, List, Union

from platypush.message import Mapping


class Pin(Mapping):
    """
    This class models the configuration for the PIN of a device.
    """
    def __init__(self, number: int, name: Optional[str] = None, pwm: bool = False, pull_up: bool = False):
        super().__init__(number=number, name=name, pwm=pwm, pull_up=pull_up)


class Device(Mapping):
    """
    This class models the properties of a configured ESP device.
    """
    def __init__(self, host: str, port: int = 8266, password: Optional[str] = None,
                 name: Optional[str] = None, pins: List[Union[Pin, dict]] = None):
        pins = [
            pin if isinstance(pin, Pin) else Pin(**pin)
            for pin in (pins or [])
        ]

        super().__init__(host=host, port=port, password=password, pins=pins, name=name)
        self._pin_by_name = {pin['name']: pin for pin in self['pins'] if pin['name']}
        self._pin_by_number = {pin['number']: pin for pin in self['pins']}

    def get_pin(self, pin) -> int:
        try:
            return int(pin)
        except ValueError:
            pass

        assert pin in self._pin_by_name, 'No such PIN configured: {}'.format(pin)
        return self._pin_by_name[pin]['number']


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import json
import os
from typing import Optional, Dict, Any

from platypush.context import get_bus, get_plugin
from platypush.message.event.assistant import ConversationStartEvent, \
    ConversationEndEvent, SpeechRecognizedEvent, VolumeChangedEvent, \
    ResponseEvent

from platypush.message.event.google import GoogleDeviceOnOffEvent

from platypush.plugins import action
from platypush.plugins.assistant import AssistantPlugin


class AssistantGooglePushtotalkPlugin(AssistantPlugin):
    """
    Plugin for the Google Assistant push-to-talk API.

    Triggers:

        * :class:`platypush.message.event.assistant.ConversationStartEvent`
            when a new conversation starts
        * :class:`platypush.message.event.assistant.SpeechRecognizedEvent`
            when a new voice command is recognized
        * :class:`platypush.message.event.assistant.ConversationEndEvent`
            when a new conversation ends

    Requires:

        * **tenacity** (``pip install tenacity``)
        * **google-assistant-sdk** (``pip install google-assistant-sdk[samples]``)
    """

    api_endpoint = 'embeddedassistant.googleapis.com'
    grpc_deadline = 60 * 3 + 5
    device_handler = None

    def __init__(self,
                 credentials_file=os.path.join(
                     os.path.expanduser('~'), '.config',
                     'google-oauthlib-tool', 'credentials.json'),
                 device_config=os.path.join(
                     os.path.expanduser('~'), '.config', 'googlesamples-assistant',
                     'device_config.json'),
                 language='en-US',
                 play_response=True,
                 tts_plugin=None,
                 tts_args=None,
                 **kwargs):
        """
        :param credentials_file: Path to the Google OAuth credentials file
            (default: ~/.config/google-oauthlib-tool/credentials.json).
            See https://developers.google.com/assistant/sdk/guides/library/python/embed/install-sample#generate_credentials
            for instructions to get your own credentials file.
        :type credentials_file: str

        :param device_config: Path to device_config.json. Register your device
            (see https://developers.google.com/assistant/sdk/guides/library/python/embed/register-device)
            and create a project, then run the pushtotalk.py script from
            googlesamples to create your device_config.json
        :type device_config: str

        :param language: Assistant language (default: en-US)
        :type language: str

        :param play_response: If True (default) then the plugin will play the assistant response upon processed
            response. Otherwise nothing will be played - but you may want to handle the ``ResponseEvent`` manually.
        :type play_response: bool

        :param tts_plugin: Optional text-to-speech plugin to be used to process response text.
        :type tts_plugin: str

        :param tts_args: Optional arguments for the TTS plugin ``say`` method.
        :type tts_args: dict
        """

        import googlesamples.assistant.grpc.audio_helpers as audio_helpers
        super().__init__(**kwargs)

        self.audio_sample_rate = audio_helpers.DEFAULT_AUDIO_SAMPLE_RATE
        self.audio_sample_width = audio_helpers.DEFAULT_AUDIO_SAMPLE_WIDTH
        self.audio_iter_size = audio_helpers.DEFAULT_AUDIO_ITER_SIZE
        self.audio_block_size = audio_helpers.DEFAULT_AUDIO_DEVICE_BLOCK_SIZE
        self.audio_flush_size = audio_helpers.DEFAULT_AUDIO_DEVICE_FLUSH_SIZE

        self.language = language
        self.credentials_file = credentials_file
        self.device_config = device_config
        self.play_response = play_response
        self.tts_plugin = tts_plugin
        self.tts_args = tts_args or {}
        self.assistant = None
        self.interactions = []

        with open(self.device_config) as f:
            device = json.load(f)
            self.device_id = device['id']
            self.device_model_id = device['model_id']

        # Load OAuth 2.0 credentials.
        try:
            from google.oauth2.credentials import Credentials
            from google.auth.transport.requests import Request

            with open(self.credentials_file, 'r') as f:
                self.credentials = Credentials(token=None, **json.load(f))
                self.http_request = Request()
                self.credentials.refresh(self.http_request)
        except Exception as ex:
            self.logger.error('Error loading credentials: %s', str(ex))
            self.logger.error('Run google-oauthlib-tool to initialize '
                              'new OAuth 2.0 credentials.')
            raise

        self.grpc_channel = None
        self.conversation_stream = None

    def _init_assistant(self):
        import googlesamples.assistant.grpc.audio_helpers as audio_helpers
        from google.auth.transport.grpc import secure_authorized_channel

        self.interactions = []

        # Create an authorized gRPC channel.
        self.grpc_channel = secure_authorized_channel(self.credentials, self.http_request, self.api_endpoint)
        self.logger.info('Connecting to {}'.format(self.api_endpoint))

        # Configure audio source and sink.
        audio_device = None
        audio_source = audio_device = (
            audio_device or audio_helpers.SoundDeviceStream(
                sample_rate=self.audio_sample_rate,
                sample_width=self.audio_sample_width,
                block_size=self.audio_block_size,
                flush_size=self.audio_flush_size
            )
        )

        audio_sink = audio_device = (
            audio_device or audio_helpers.SoundDeviceStream(
                sample_rate=self.audio_sample_rate,
                sample_width=self.audio_sample_width,
                block_size=self.audio_block_size,
                flush_size=self.audio_flush_size
            )
        )

        # Create conversation stream with the given audio source and sink.
        self.conversation_stream = audio_helpers.ConversationStream(
            source=audio_source,
            sink=audio_sink,
            iter_size=self.audio_iter_size,
            sample_width=self.audio_sample_width,
        )

        self._install_device_handlers()

    def on_conversation_start(self):
        """ Conversation start handler """
        def handler():
            get_bus().post(ConversationStartEvent(assistant=self))

        return handler

    def on_conversation_end(self):
        """ Conversation end handler """
        def handler(with_follow_on_turn):
            get_bus().post(ConversationEndEvent(assistant=self, with_follow_on_turn=with_follow_on_turn))

        return handler

    def on_speech_recognized(self):
        """ Speech recognized handler """
        def handler(phrase):
            get_bus().post(SpeechRecognizedEvent(assistant=self, phrase=phrase))
            self.interactions.append({'request': phrase})

        return handler

    def on_volume_changed(self):
        """ Volume changed event """
        def handler(volume):
            get_bus().post(VolumeChangedEvent(assistant=self, volume=volume))

        return handler

    def on_response(self):
        """ Response handler """
        def handler(response):
            get_bus().post(ResponseEvent(assistant=self, response_text=response))

            if not self.interactions:
                self.interactions.append({'response': response})
            else:
                self.interactions[-1]['response'] = response

            if self.tts_plugin:
                tts = get_plugin(self.tts_plugin)
                tts.say(response, **self.tts_args)

        return handler

    @action
    def start_conversation(self, *args, language: Optional[str] = None, tts_plugin: Optional[str] = None,
                           tts_args: Optional[Dict[str, Any]] = None, **kwargs):
        """
        Start a conversation

        :param language: Language code override (default: default configured language).
        :param tts_plugin: Optional text-to-speech plugin to be used for rendering text.
        :param tts_args: Optional arguments for the TTS plugin say method.

        :returns: A list of the interactions that happen within the conversation.

        ..code-block:: json

            [
                {
                    "request": "request 1",
                    "response": "response 1"

                },

                {
                    "request": "request 2",
                    "response": "response 2"

                }
            ]

        """

        from platypush.plugins.assistant.google.lib import SampleAssistant

        self.tts_plugin = tts_plugin
        self.tts_args = tts_args
        language = language or self.language
        play_response = False if self.tts_plugin else self.play_response

        self._init_assistant()
        self.on_conversation_start()

        with SampleAssistant(language_code=language,
                             device_model_id=self.device_model_id,
                             device_id=self.device_id,
                             conversation_stream=self.conversation_stream,
                             display=None,
                             channel=self.grpc_channel,
                             deadline_sec=self.grpc_deadline,
                             play_response=play_response,
                             device_handler=self.device_handler,
                             on_conversation_start=self.on_conversation_start(),
                             on_conversation_end=self.on_conversation_end(),
                             on_volume_changed=self.on_volume_changed(),
                             on_response=self.on_response(),
                             on_speech_recognized=self.on_speech_recognized()) as self.assistant:
            continue_conversation = True

            while continue_conversation:
                try:
                    continue_conversation = self.assistant.assist()
                except Exception as e:
                    self.logger.warning('Unhandled assistant exception: {}'.format(str(e)))
                    self.logger.exception(e)
                    self._init_assistant()

        return self.interactions

    @action
    def stop_conversation(self):
        if self.assistant:
            self.assistant.play_response = False

            if self.conversation_stream:
                self.conversation_stream.stop_playback()
                self.conversation_stream.stop_recording()

            get_bus().post(ConversationEndEvent(assistant=self))

    @action
    def set_mic_mute(self, muted: bool = True):
        """
        Programmatically mute/unmute the microphone.

        :param muted: Set to True or False.
        """
        if not self.conversation_stream:
            self.logger.warning('The assistant is not running')
            return

        if muted:
            self.conversation_stream.stop_recording()
        else:
            self.conversation_stream.start_recording()

    def _install_device_handlers(self):
        import googlesamples.assistant.grpc.device_helpers as device_helpers
        self.device_handler = device_helpers.DeviceRequestHandler(self.device_id)

        @self.device_handler.command('action.devices.commands.OnOff')
        def handler(on):
            get_bus().post(GoogleDeviceOnOffEvent(
                device_id=self.device_id,
                device_model_id=self.device_model_id,
                on=on))


# vim:sw=4:ts=4:et:



from platypush.backend.sensor import SensorBackend
from platypush.context import get_plugin


class SensorDistanceBackend(SensorBackend):
    """
    Backend to poll a distance sensor.

    Requires:

        * ``RPi.GPIO`` (``pip install RPi.GPIO``)
        * The :mod:`platypush.plugins.gpio.sensor.distance` plugin configured
    """

    def get_measurement(self):
        """ get_measurement implementation """
        plugin = get_plugin('gpio.sensor.distance')
        return {"distance": plugin.get_data().output }


# vim:sw=4:ts=4:et:





from platypush.backend.sensor import SensorBackend


class SensorMotionPwm3901Backend(SensorBackend):
    """
    Backend to poll an `PWM3901 <https://github.com/pimoroni/pmw3901-python>`_
    optical flow and motion sensor

    Requires:

        * ``pwm3901`` (``pip install pwm3901``)
    """

    _default_poll_seconds = 0.01

    def __init__(self, **kwargs):
        if 'poll_seconds' not in kwargs:
            # noinspection PyTypeChecker
            kwargs['poll_seconds'] = self._default_poll_seconds

        super().__init__(plugin='gpio.sensor.motion.pwm3901', **kwargs)


# vim:sw=4:ts=4:et:



"""
Controller for Plex content on a Chromecast device
"""

from pychromecast.controllers import BaseController


MESSAGE_TYPE = 'type'

TYPE_PLAY = "PLAY"
TYPE_PAUSE = "PAUSE"
TYPE_STOP = "STOP"

class PlexController(BaseController):
    """ Controller to interact with Plex namespace. """

    def __init__(self):
        super(PlexController, self).__init__(
            "urn:x-cast:plex", "9AC194DC")
        self.app_id="9AC194DC"
        self.namespace="urn:x-cast:plex"
        self.request_id = 0

    def stop(self):
        """ Send stop command. """
        self.namespace = "urn:x-cast:plex"
        self.request_id += 1
        self.send_message({MESSAGE_TYPE: TYPE_STOP})

    def pause(self):
        """ Send pause command. """
        self.namespace = "urn:x-cast:plex"
        self.request_id += 1
        self.send_message({MESSAGE_TYPE: TYPE_PAUSE})

    def play(self):
        """ Send play command. """
        self.namespace = "urn:x-cast:plex"
        self.request_id += 1
        self.send_message({MESSAGE_TYPE: TYPE_PLAY})


    def play_media(self,item,server,medtype="LOAD"):
        def app_launched_callback():
                self.set_load(item,server,medtype)

        receiver_ctrl = self._socket_client.receiver_controller
        if receiver_ctrl.status.app_id != self.app_id:
            receiver_ctrl.launch_app(self.app_id,
                                     callback_function=app_launched_callback)

    def set_load(self,item,server,medtype="LOAD"):
        transient_token = server.query("/security/token?type=delegation&scope=all").attrib.get('token')
        playqueue = server.createPlayQueue(item).playQueueID
        self.request_id += 1
        address = server.url('').split(":")[1][2:]
        self.namespace="urn:x-cast:com.google.cast.media"
        msg = {
                "type": medtype,
                "requestId": self.request_id,
                "sessionId": "python_player",
                "autoplay": True,
                "currentTime": 0,
                "media":{
                        "contentId": item.key,
                        "streamType": "BUFFERED",
                        "contentType": "video",
                        "customData": {
                                "offset": 0,
                                "server": {
                                        "machineIdentifier": server.machineIdentifier,
                                        "transcoderVideo": True,
                                        "transcoderVideoRemuxOnly": False,
                                        "transcoderAudio": True,
                                        "version": "1.3.3.3148",
                                        "myPlexSubscription": False,
                                        "isVerifiedHostname": True,
                                        "protocol": "https",
                                        "address": address,
                                        "port": "32400",
                                        "accessToken": transient_token,
                                },
                                "user": {"username": server.myPlexUsername},
                                "containerKey": "/playQueues/{}?own=1&window=200".format(playqueue),
                        },
                }
        }
        self.send_message(msg, inc_session_id=True)

# vim:sw=4:ts=4:et:




import copy
import json
import logging
import re
import threading

from platypush.config import Config
from platypush.message.event import Event, EventMatchResult
from platypush.message.request import Request
from platypush.procedure import Procedure
from platypush.utils import get_event_class_by_type, set_thread_name

logger = logging.getLogger(__name__)


def parse(msg):
    """ Builds a dict given another dictionary or
        a JSON UTF-8 encoded string/bytearray """

    if isinstance(msg, bytes) or isinstance(msg, bytearray):
        msg = msg.decode('utf-8')
    if isinstance(msg, str):
        try:
            msg = json.loads(msg.strip())
        except:
            logger.warning('Invalid JSON message: {}'.format(msg))
            return None

    return msg


class EventCondition(object):
    """ Event hook condition class """

    def __init__(self, type=Event.__class__, priority=None, **kwargs):
        """
        Rule constructor.
        Params:
            type   -- Class of the event to be built
            kwargs -- Fields rules as a key-value (e.g. source_button=btn_id
                or recognized_phrase='Your phrase')
        """

        self.type = type
        self.args = {}
        self.parsed_args = {}
        self.priority = priority

        for (key, value) in kwargs.items():
            # TODO So far we only allow simple value match. If value is a dict
            # instead, we should allow more a sophisticated attribute matching,
            # e.g. or conditions, in, and other operators.
            self.args[key] = value

    @classmethod
    def build(cls, rule):
        """ Builds a rule given either another EventRule, a dictionary or
            a JSON UTF-8 encoded string/bytearray """

        if isinstance(rule, cls): return rule
        else: rule = parse(rule)
        assert isinstance(rule, dict)

        type = get_event_class_by_type(
            rule.pop('type') if 'type' in rule else 'Event')

        args = {}
        for (key, value) in rule.items():
            args[key] = value

        return cls(type=type, **args)


class EventAction(Request):
    """ Event hook action class. It is a special type of runnable request
        whose fields can be configured later depending on the event context """

    def __init__(self, target=None, action=None, **args):
        if target is None: target=Config.get('device_id')
        args_copy = copy.deepcopy(args)
        super().__init__(target=target, action=action, **args_copy)


    @classmethod
    def build(cls, action):
        action = super().parse(action)
        action['origin'] = Config.get('device_id')

        if 'target' not in action:
            action['target'] = action['origin']

        token = Config.get('token')
        if token:
            action['token'] = token

        return super().build(action)


class EventHook(object):
    """ Event hook class. It consists of one conditionss and
        one or multiple actions to be executed """

    def __init__(self, name, priority=None, condition=None, actions=[]):
        """ Construtor. Takes a name, a EventCondition object and an event action
            procedure as input. It may also have a priority attached
            as a positive number. If multiple hooks match against an event,
            only the ones that have either the maximum match score or the
            maximum pre-configured priority will be run. """

        self.name = name
        self.condition = EventCondition.build(condition or {})
        self.actions = actions
        self.priority = priority or 0
        self.condition.priority = self.priority


    @classmethod
    def build(cls, name, hook):
        """ Builds a rule given either another EventRule, a dictionary or
            a JSON UTF-8 encoded string/bytearray """

        if isinstance(hook, cls): return hook
        else: hook = parse(hook)
        assert isinstance(hook, dict)

        condition = EventCondition.build(hook['if']) if 'if' in hook else None
        actions = []
        priority = hook['priority'] if 'priority' in hook else None
        condition.priority = priority

        if 'then' in hook:
            if isinstance(hook['then'], list):
                actions = hook['then']
            else:
                actions = [hook['then']]

        actions = Procedure.build(name=name+'__Hook', requests=actions, _async=False)
        return cls(name=name, condition=condition, actions=actions, priority=priority)


    def matches_event(self, event):
        """ Returns an EventMatchResult object containing the information
            about the match between the event and this hook """

        return event.matches_condition(self.condition)


    def run(self, event):
        """ Checks the condition of the hook against a particular event and
            runs the hook actions if the condition is met """

        def _thread_func(result):
            set_thread_name('Event-' + self.name)
            self.actions.execute(event=event, **result.parsed_args)

        result = self.matches_event(event)
        token = Config.get('token')

        if result.is_match:
            logger.info('Running hook {} triggered by an event'.format(self.name))
            threading.Thread(target=_thread_func,
                             name='Event-' + self.name,
                             args=(result,)).start()


# vim:sw=4:ts=4:et:




import time

from typing import Optional

import todoist
import todoist.managers.items

from platypush.plugins import Plugin, action
from platypush.message.response.todoist import TodoistUserResponse, TodoistProjectsResponse, TodoistItemsResponse, \
    TodoistFiltersResponse, TodoistLiveNotificationsResponse, TodoistCollaboratorsResponse, TodoistNotesResponse, \
    TodoistProjectNotesResponse


class TodoistPlugin(Plugin):
    """
    Todoist integration.

    Requires:

        * **todoist-python** (``pip install todoist-python``)

    You'll also need a Todoist token. You can get it `here <https://todoist.com/prefs/integrations>`.
    """

    _sync_timeout = 60.0

    def __init__(self, api_token: str, **kwargs):
        """
        :param api_token: Todoist API token. You can get it `here <https://todoist.com/prefs/integrations>`.
        """

        super().__init__(**kwargs)
        self.api_token = api_token
        self._api = None
        self._last_sync_time = None

    def _get_api(self) -> todoist.TodoistAPI:
        if not self._api:
            self._api = todoist.TodoistAPI(self.api_token)

        if not self._last_sync_time or time.time() - self._last_sync_time > self._sync_timeout:
            self._api.sync()

        return self._api

    @action
    def get_user(self) -> TodoistUserResponse:
        """
        Get logged user info.
        """
        api = self._get_api()
        return TodoistUserResponse(**api.state['user'])

    @action
    def get_projects(self) -> TodoistProjectsResponse:
        """
        Get list of Todoist projects.
        """
        api = self._get_api()
        return TodoistProjectsResponse(api.state['projects'])

    @action
    def get_items(self) -> TodoistItemsResponse:
        """
        Get list of Todoist projects.
        """
        api = self._get_api()
        return TodoistItemsResponse(api.state['items'])

    @action
    def get_filters(self) -> TodoistFiltersResponse:
        """
        Get list of Todoist filters.
        """
        api = self._get_api()
        return TodoistFiltersResponse(api.state['filters'])

    @action
    def get_live_notifications(self) -> TodoistLiveNotificationsResponse:
        """
        Get list of Todoist live notifications.
        """
        api = self._get_api()
        return TodoistLiveNotificationsResponse(api.state['live_notifications'])

    @action
    def get_collaborators(self) -> TodoistCollaboratorsResponse:
        """
        Get list of collaborators.
        """
        api = self._get_api()
        return TodoistCollaboratorsResponse(api.state['collaborators'])

    @action
    def get_notes(self) -> TodoistNotesResponse:
        """
        Get list of Todoist notes.
        """
        api = self._get_api()
        return TodoistNotesResponse(api.state['notes'])

    @action
    def get_project_notes(self) -> TodoistProjectNotesResponse:
        """
        Get list of Todoist project notes.
        """
        api = self._get_api()
        return TodoistProjectNotesResponse(api.state['project_notes'])

    @action
    def add_item(self, content: str, project_id: Optional[int] = None, **kwargs):
        """
        Add a new item.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        item = manager.add(content, project_id=project_id, **kwargs)
        api.commit()
        return item.data

    @action
    def delete_item(self, item_id: int):
        """
        Delete an item by id.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.delete(item_id)
        api.commit()

    @action
    def update_item(self, item_id: int, **kwargs):
        """
        Update an item by id.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.update(item_id, **kwargs)
        api.commit()

    @action
    def complete_item(self, item_id: int):
        """
        Mark an item as done.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.complete(item_id)
        api.commit()

    @action
    def uncomplete_item(self, item_id: int):
        """
        Mark an item as not done.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.uncomplete(item_id)
        api.commit()

    @action
    def archive(self, item_id: int):
        """
        Archive an item by id.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.archive(item_id)
        api.commit()

    @action
    def unarchive(self, item_id: int):
        """
        Un-archive an item by id.
        """
        api = self._get_api()
        manager = todoist.managers.items.ItemsManager(api=api)
        manager.unarchive(item_id)
        api.commit()

    @action
    def sync(self):
        """
        Sync/update info with the remote server.
        """
        api = self._get_api()
        api.sync()


# vim:sw=4:ts=4:et:



import subprocess

from platypush.plugins import Plugin, action


class ShellPlugin(Plugin):
    """
    Plugin to run custom shell commands.
    """

    @action
    def exec(self, cmd, ignore_errors=False):
        """
        Execute a command.

        :param cmd: Command to execute
        :type cmd: str

        :param ignore_errors: If set, then any errors in the command execution will be ignored. Otherwise a RuntimeError will be thrown (default value: False)
        :returns: A response object where the ``output`` field will contain the command output as a string, and the ``errors`` field will contain whatever was sent to stderr.
        """

        try:
            return subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, shell=True).decode('utf-8')
        except subprocess.CalledProcessError as e:
            if ignore_errors:
                self.logger.warning('Command {} failed with error: {}'.format(
                    cmd, e.output.decode('utf-8')))
            else:
                raise RuntimeError(e.output.decode('utf-8'))


# vim:sw=4:ts=4:et:




import threading

from abc import ABC, abstractmethod
from queue import Queue
from typing import Type


class EndOfStream:
    pass


class Worker(ABC, threading.Thread):
    """
    Generic class for worker threads, used to split the execution of an action over multiple
    parallel instances.
    """

    def __init__(self, request_queue: Queue, response_queue=Queue, id=None):
        """
        :param request_queue: The worker will listen for messages to process over this queue
        :param response_queue: The worker will return responses over this queue
        """
        super().__init__()
        self.request_queue = request_queue
        self.response_queue = response_queue
        self._id = id

    def run(self) -> None:
        """
        The worker will run until it receives a :class:`EndOfStream` message on the queue.
        """
        while True:
            msg = self.request_queue.get()
            if isinstance(msg, EndOfStream):
                break

            try:
                ret = self.process(msg)
            except Exception as e:
                ret = e

            if ret:
                # noinspection PyArgumentList,PyCallByClass
                self.response_queue.put(ret)

    @abstractmethod
    def process(self, msg):
        """
        This method must be implemented by the derived classes.
        It will take as argument a message received over the `request_queue` and will return a value that will be
        processed by the consumer or None.

        If this function raises an exception then the exception will be pushed to the response queue and can be
        handled by the consumer.
        """
        raise NotImplementedError('Must be implemented by a derived class')

    def end_stream(self) -> None:
        """
        This method will be called when we have no more messages to send to the worker. A special
        `EndOfStream` object will be sent on the `request_queue`.
        """
        self.request_queue.put(EndOfStream())


class Workers:
    """
    Model for a pool of workers. Syntax:

    .. code-block:: python

        class Squarer(Worker):
            def process(self, n):
                return n * n

        workers = Workers(5, Squarer)  # Allocate 5 workers of type Squarer
        with workers:
            for n in range(100)):
                workers.put(n)

        print(workers.responses)

    """

    def __init__(self, n_workers: int, worker_type: Type[Worker], *args, **kwargs):
        """
        :param n_workers: Number of workers
        :param worker_type: Type of the workers that will be instantiated. Must be a subclass of :class:`Worker`.
        :param args: Extra args to pass to the `worker_type` constructor
        :param kwargs: Extra kwargs to pass to the `worker_type` constructor
        """
        self.request_queue = Queue()
        self.response_queue = Queue()
        # noinspection PyArgumentList
        self._workers = [worker_type(self.request_queue, self.response_queue, id=i, *args, **kwargs)
                         for i in range(n_workers)]
        self.responses = []

    def start(self):
        for wrk in self._workers:
            wrk.start()

    def put(self, msg) -> None:
        """
        Put a message on the `request_queue`
        """
        self.request_queue.put(msg)

    def wait(self) -> list:
        """
        Wait for the termination of all the workers

        :ret: A list containing the processed responses
        """
        while self._workers:
            wrk = self._workers.pop()
            wrk.join()

        while not self.response_queue.empty():
            self.responses.append(self.response_queue.get())

        return self.responses

    def end_stream(self):
        """
        Mark the termination of the stream by sending an :class:`EndOfStream` message on the `request_queue` for
        each of the running workers.
        """
        for wrk in self._workers:
            wrk.end_stream()

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_stream()
        self.wait()


# vim:sw=4:ts=4:et:



import json
import logging
import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message import Message
from platypush.message.event.kafka import KafkaMessageEvent


class KafkaBackend(Backend):
    """
    Backend to interact with an Apache Kafka (https://kafka.apache.org/)
    streaming platform, send and receive messages.

    Requires:

        * **kafka** (``pip install kafka-python``)
    """

    _conn_retry_secs = 5

    def __init__(self, server='localhost:9092', topic='platypush', **kwargs):
        """
        :param server: Kafka server name or address + port (default: ``localhost:9092``)
        :type server: str

        :param topic: (Prefix) topic to listen to (default: platypush). The Platypush device_id (by default the hostname) will be appended to the topic (the real topic name will e.g. be "platypush.my_rpi")
        :type topic: str
        """

        super().__init__(**kwargs)

        self.server = server
        self.topic_prefix = topic
        self.topic = self._topic_by_device_id(self.device_id)
        self.producer = None

        # Kafka can be veryyyy noisy
        logging.getLogger('kafka').setLevel(logging.ERROR)

    def _on_record(self, record):
        if record.topic != self.topic: return
        msg = record.value.decode('utf-8')
        is_platypush_message = False

        try:
            msg = Message.build(msg)
            is_platypush_message = True
        except:
            pass

        self.logger.info('Received message on Kafka backend: {}'.format(msg))

        if is_platypush_message:
            self.on_message(msg)
        else:
            self.on_message(KafkaMessageEvent(msg=msg))

    def _topic_by_device_id(self, device_id):
        return '{}.{}'.format(self.topic_prefix, device_id)

    def send_message(self, msg):
        target = msg.target
        kafka_plugin = get_plugin('kafka')
        kafka_plugin.send_message(msg=msg,
                                  topic=self._topic_by_device_id(target),
                                  server=self.server)

    def on_stop(self):
        try:
            if self.producer:
                self.producer.flush()
                self.producer.close()

            if self.consumer:
                self.consumer.close()
        except Exception as e:
            self.logger.warning('Exception occurred while closing Kafka connection')
            self.logger.exception(e)

    def run(self):
        from kafka import KafkaConsumer
        super().run()

        self.consumer = KafkaConsumer(self.topic, bootstrap_servers=self.server)
        self.logger.info('Initialized kafka backend - server: {}, topic: {}'
                     .format(self.server, self.topic))

        try:
            for msg in self.consumer:
                self._on_record(msg)
                if self.should_stop(): break
        except Exception as e:
            self.logger.warning('Kafka connection error, reconnecting in {} seconds'.
                            format(self._conn_retry_secs))
            self.logger.exception(e)
            time.sleep(self._conn_retry_secs)

# vim:sw=4:ts=4:et:




import json
import time

from platypush.message import Message


class Response(Message):
    """ Response message class """

    def __init__(self, target=None, origin=None, id=None, output=None, errors=None,
                 timestamp=None, disable_logging=False):
        """
        Params:
            target -- Target [String]
            origin -- Origin [String]
            output -- Output [String]
            errors -- Errors [List of strings or exceptions]
                id -- Message ID this response refers to
         timestamp -- Message timestamp [Float]
        """

        super().__init__(timestamp=timestamp)
        self.target = target
        self.output = self._parse_msg(output)
        self.errors = self._parse_msg(errors or [])
        self.origin = origin
        self.id = id
        self.disable_logging = disable_logging

    def is_error(self):
        """ Returns True if the response has errors """
        return len(self.errors) != 0

    @classmethod
    def _parse_msg(cls, msg):
        if isinstance(msg, bytes) or isinstance(msg, bytearray):
            msg = msg.decode('utf-8')
        if isinstance(msg, str):
            try:
                msg = json.loads(msg.strip())
            except ValueError:
                pass

        return msg

    @classmethod
    def build(cls, msg):
        msg = super().parse(msg)
        args = {
            'target': msg['target'],
            'output': msg['response']['output'],
            'errors': msg['response']['errors'],
            'timestamp': msg['_timestamp'] if '_timestamp' in msg else time.time(),
            'disable_logging': msg.get('_disable_logging', False),
        }

        if 'id' in msg:
            args['id'] = msg['id']
        if 'origin' in msg:
            args['origin'] = msg['origin']

        return cls(**args)

    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """
        output = self.output if self.output is not None else {
            'success': True if not self.errors else False
        }

        response_dict = {
            'id': self.id,
            'type': 'response',
            'target': self.target if hasattr(self, 'target') else None,
            'origin': self.origin if hasattr(self, 'origin') else None,
            '_timestamp': self.timestamp,
            'response': {
                'output': output,
                'errors': self.errors,
            },
        }

        if self.disable_logging:
            response_dict['_disable_logging'] = self.disable_logging

        return json.dumps(response_dict, cls=self.Encoder)


# vim:sw=4:ts=4:et:



import os
import stat

# noinspection PyPackageRequirements
from PyOBEX import requests, responses, headers
# noinspection PyPackageRequirements
from PyOBEX.server import BrowserServer

from platypush.backend.bluetooth import BluetoothBackend
from platypush.message.event.bluetooth import BluetoothFileGetRequestEvent


class BluetoothFileserverBackend(BluetoothBackend, BrowserServer):
    """
    Bluetooth OBEX file server.
    Enable it to allow bluetooth devices to browse files on this machine.

    If you run platypush as a non-root user (and you should) then you to change the group owner of the
    service discovery protocol file (/var/run/sdp) and add your user to that group. See
    `here <https://stackoverflow.com/questions/34599703/rfcomm-bluetooth-permission-denied-error-raspberry-pi>`_
    for details.

    Requires:

        * **pybluez** (``pip install pybluez``)
        * **pyobex** (``pip install git+https://github.com/BlackLight/PyOBEX``)

    """

    def __init__(self, port: int, address: str = '', directory: str = os.path.expanduser('~'),
                 whitelisted_addresses: list = None, **kwargs):
        """
        :param port: Bluetooth listen port
        :param address: Bluetooth address to bind the server to (default: any)
        :param directory: Directory to share (default: HOME directory)
        :param whitelisted_addresses: If set then only accept connections from the listed device addresses
        """
        BluetoothBackend.__init__(self, address=address, port=port, directory=directory,
                                  whitelisted_addresses=whitelisted_addresses, **kwargs)

        if not os.path.isdir(self.directory):
            raise FileNotFoundError(self.directory)

    def process_request(self, socket, request, *address):
        if isinstance(request, requests.Get):
            self.bus.post(BluetoothFileGetRequestEvent(address=address[0], port=address[1]))
            self.get(socket, request)
        else:
            super().process_request(socket, request, *address)

    def get(self, socket, request):
        name = ""
        req_type = ""

        for header in request.header_data:
            if isinstance(header, headers.Name):
                name = header.decode().strip("\x00")
                self.logger.info("Receiving request for {}".format(name))
            elif isinstance(header, headers.Type):
                req_type = header.decode().strip("\x00")
                self.logger.info("Request type: {}".format(req_type))

        path = os.path.abspath(os.path.join(self.directory, name))

        if os.path.isdir(path) or req_type == "x-obex/folder-listing":
            if path.startswith(self.directory):
                filelist = os.listdir(path)
                s = '<?xml version="1.0"?>\n<folder-listing>\n'

                for i in filelist:
                    objpath = os.path.join(path, i)
                    if os.path.isdir(objpath):
                        s += '  <folder name="{}" created="{}" />'.format(i, os.stat(objpath)[stat.ST_CTIME])
                    else:
                        s += '  <file name="{}" created="{}" size="{}" />'.format(
                            i, os.stat(objpath)[stat.ST_CTIME], os.stat(objpath)[stat.ST_SIZE])

                s += "</folder-listing>\n"
                self.logger.debug('Bluetooth get XML output:\n' + s)

                response = responses.Success()
                response_headers = [headers.Name(name.encode("utf8")),
                                    headers.Length(len(s)),
                                    headers.Body(s.encode("utf8"))]
                BrowserServer.send_response(self, socket, response, response_headers)
            else:
                self._reject(socket)
        else:
            self._reject(socket)


# vim:sw=4:ts=4:et:



import ast
import statistics
import json
import time

from threading import Thread, Lock

from platypush.context import get_backend
from platypush.plugins import Plugin, action

data_throttler_lock = None


class AdafruitIoPlugin(Plugin):
    """
    This plugin allows you to interact with the Adafruit IO
    <https://io.adafruit.com>, a cloud-based message queue and storage.
    You can send values to feeds on your Adafruit IO account and read the
    values of those feeds as well through any device.

    Requires:

        * **adafruit-io** (``pip install adafruit-io``)
        * Redis server running and Redis backend configured if you want to enable throttling

    Some example usages::

        # Send the temperature value for a connected sensor to the "temperature" feed
        {
            "type": "request",
            "action": "adafruit.io.send",
            "args": {
                "feed": "temperature",
                "value": 25.0
            }
        }

        # Receive the most recent temperature value
        {
            "type": "request",
            "action": "adafruit.io.receive",
            "args": {
                "feed": "temperature"
            }
        }
    """

    _DATA_THROTTLER_QUEUE = 'platypush/adafruit.io'

    def __init__(self, username, key, throttle_seconds=None, **kwargs):
        """
        :param username: Your Adafruit username
        :type username: str

        :param key: Your Adafruit IO key
        :type key: str

        :param throttle_seconds: If set, then instead of sending the values directly over ``send`` the plugin will
            first collect all the samples within the specified period and then dispatch them to Adafruit IO.
            You may want to set it if you have data sources providing a lot of data points and you don't want to hit
            the throttling limitations of Adafruit.
        :type throttle_seconds: float
        """

        from Adafruit_IO import Client
        global data_throttler_lock
        super().__init__(**kwargs)

        self._username = username
        self._key = key
        self.aio = Client(username=username, key=key)
        self.throttle_seconds = throttle_seconds

        if not data_throttler_lock:
            data_throttler_lock = Lock()

        if self.throttle_seconds and not data_throttler_lock.locked():
            self._get_redis()
            self.logger.info('Starting Adafruit IO throttler thread')
            data_throttler_lock.acquire(False)
            self.data_throttler = Thread(target=self._data_throttler())
            self.data_throttler.start()

    @staticmethod
    def _get_redis():
        from redis import Redis

        redis_args = get_backend('redis').redis_args
        redis_args['socket_timeout'] = 1
        return Redis(**redis_args)

    def _data_throttler(self):
        from redis.exceptions import TimeoutError as QueueTimeoutError

        def run():
            redis = self._get_redis()
            last_processed_batch_timestamp = None
            data = {}

            try:
                while True:
                    try:
                        new_data = ast.literal_eval(
                            redis.blpop(self._DATA_THROTTLER_QUEUE)[1].decode('utf-8'))

                        for (key, value) in new_data.items():
                            data.setdefault(key, []).append(value)
                    except QueueTimeoutError:
                        pass

                    if data and (last_processed_batch_timestamp is None or
                                 time.time() - last_processed_batch_timestamp >= self.throttle_seconds):
                        last_processed_batch_timestamp = time.time()
                        self.logger.info('Processing feeds batch for Adafruit IO')

                        for (feed, values) in data.items():
                            if values:
                                value = statistics.mean(values)

                                try:
                                    self.send(feed, value, enqueue=False)
                                except ThrottlingError:
                                    self.logger.warning('Adafruit IO throttling threshold hit, taking a nap ' +
                                                        'before retrying')
                                    time.sleep(self.throttle_seconds)

                        data = {}
            except Exception as e:
                self.logger.exception(e)

        return run

    @action
    def send(self, feed, value, enqueue=True):
        """
        Send a value to an Adafruit IO feed

        :param feed: Feed name
        :type feed: str

        :param value: Value to send
        :type value: Numeric or string

        :param enqueue: If throttle_seconds is set, this method by default will append values to the throttling queue
            to be periodically flushed instead of sending the message directly. In such case, pass enqueue=False to
            override the behaviour and send the message directly instead.
        :type enqueue: bool
        """

        if not self.throttle_seconds or not enqueue:
            # If no throttling is configured, or enqueue is false then send the value directly to Adafruit
            self.aio.send(feed, value)
        else:
            # Otherwise send it to the Redis queue to be picked up by the throttler thread
            redis = self._get_redis()
            redis.rpush(self._DATA_THROTTLER_QUEUE, json.dumps({feed: value}))

    @action
    def send_location_data(self, feed, lat, lon, ele, value):
        """
        Send location data to an Adafruit IO feed

        :param feed: Feed name
        :type feed: str

        :param lat: Latitude
        :type lat: float

        :param lon: Longitude
        :type lon: float

        :param ele: Elevation
        :type ele: float

        :param value: Value to send
        :type value: Numeric or string
        """

        self.aio.send_data(feed=feed, value=value, metadata={
            'lat': lat,
            'lon': lon,
            'ele': ele,
        })

    @classmethod
    def _cast_value(cls, value):
        try:
            value = float(value)
        except ValueError:
            pass

        return value

    def _convert_data_to_dict(self, *data):
        from Adafruit_IO.model import DATA_FIELDS

        return [
            {
                attr: self._cast_value(getattr(i, attr))
                if attr == 'value' else getattr(i, attr)
                for attr in DATA_FIELDS if getattr(i, attr) is not None
            } for i in data
        ]

    @action
    def receive(self, feed, limit=1):
        """
        Receive data from the specified Adafruit IO feed

        :param feed: Feed name
        :type feed: str

        :param limit: Maximum number of data points to be returned. If None,
            all the values in the feed will be returned. Default: 1 (return most
            recent value)
        :type limit: int
        """

        if limit == 1:
            values = self._convert_data_to_dict(self.aio.receive(feed))
            return values[0] if values else None

        values = self._convert_data_to_dict(*self.aio.data(feed))
        return values[:limit] if limit else values

    @action
    def receive_next(self, feed):
        """
        Receive the next unprocessed data point from a feed

        :param feed: Feed name
        :type feed: str
        """

        values = self._convert_data_to_dict(self.aio.receive_next(feed))
        return values[0] if values else None

    @action
    def receive_previous(self, feed):
        """
        Receive the last processed data point from a feed

        :param feed: Feed name
        :type feed: str
        """

        values = self._convert_data_to_dict(self.aio.receive_previous(feed))
        return values[0] if values else None

    @action
    def delete(self, feed, data_id):
        """
        Delete a data point from a feed

        :param feed: Feed name
        :type feed: str

        :param data_id: Data point ID to remove
        :type data_id: str
        """

        self.aio.delete(feed, data_id)


# vim:sw=4:ts=4:et:



import os

from platypush.backend import Backend
from platypush.message.event.path import PathCreateEvent, PathDeleteEvent, \
    PathOpenEvent, PathModifyEvent, PathPermissionsChangeEvent, PathCloseEvent


class InotifyBackend(Backend):
    """
    (Linux only) This backend will listen for events on the filesystem (whether
    a file/directory on a watch list is opened, modified, created, deleted,
    closed or had its permissions changed) and will trigger a relevant event.

    Triggers:

        * :class:`platypush.message.event.path.PathCreateEvent` if a resource is created
        * :class:`platypush.message.event.path.PathOpenEvent` if a resource is opened
        * :class:`platypush.message.event.path.PathModifyEvent` if a resource is modified
        * :class:`platypush.message.event.path.PathPermissionsChangeEvent` if the permissions of a resource are changed
        * :class:`platypush.message.event.path.PathCloseEvent` if a resource is closed
        * :class:`platypush.message.event.path.PathDeleteEvent` if a resource is removed

    Requires:

        * **inotify** (``pip install inotify``)
    """

    inotify_watch = None

    def __init__(self, watch_paths=[], **kwargs):
        """
        :param watch_paths: Filesystem resources to watch for events
        :type watch_paths: str
        """

        super().__init__(**kwargs)
        self.watch_paths = set(map(
            lambda path: os.path.abspath(os.path.expanduser(path)),
            watch_paths))

    def _cleanup(self):
        if not self.inotify_watch:
            return

        for path in self.watch_paths:
            self.inotify_watch.remove_watch(path)

        self.inotify_watch = None

    def run(self):
        import inotify.adapters
        super().run()

        self.inotify_watch = inotify.adapters.Inotify()
        for path in self.watch_paths:
            self.inotify_watch.add_watch(path)

        self.logger.info('Initialized inotify file monitoring backend, monitored resources: {}'
                     .format(self.watch_paths))

        try:
            for inotify_event in self.inotify_watch.event_gen():
                if inotify_event is not None:
                    (header, inotify_types, watch_path, filename) = inotify_event
                    event = None

                    if 'IN_OPEN' in inotify_types:
                        event = PathOpenEvent(path=watch_path)
                    elif 'IN_MODIFY' in inotify_types:
                        event = PathModifyEvent(path=watch_path)
                    elif 'IN_CLOSE_WRITE' in inotify_types or 'IN_CLOSE_NOWRITE' in inotify_types:
                        event = PathCloseEvent(path=watch_path)

                    if event:
                        self.bus.post(event)
        finally:
            self._cleanup()


# vim:sw=4:ts=4:et:




"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import os

from platypush.context import get_bus
from platypush.plugins import action
from platypush.plugins.assistant import AssistantPlugin

from platypush.message.event.assistant import ConversationStartEvent, \
    ConversationEndEvent, SpeechRecognizedEvent, ResponseEvent


class AssistantEchoPlugin(AssistantPlugin):
    """
    Amazon Echo/Alexa assistant plugin.

    In order to activate the Echo service on your device follow these steps:

        1. Install avs (``pip install git+https://github.com/BlackLight/avs.git``)
        2. Run the ``alexa-auth`` script. A local webservice will start on port 3000
        3. If a browser instance doesn't open automatically then head to http://localhost:3000
        4. Log in to your Amazon account
        5. The required credentials will be stored to ~/.avs.json

    Triggers:

        * :class:`platypush.message.event.assistant.ConversationStartEvent`
            when a new conversation starts
        * :class:`platypush.message.event.assistant.SpeechRecognizedEvent`
            when a new voice command is recognized
        * :class:`platypush.message.event.assistant.ConversationEndEvent`
            when a new conversation ends

    Requires:

        * **avs** (``pip install avs``)
    """

    def __init__(self, avs_config_file: str = None, audio_device: str = 'default',
                 audio_player: str = 'default', **kwargs):
        """
        :param avs_config_file: AVS credentials file - default: ~/.avs.json. If the file doesn't exist then
            an instance of the AVS authentication service will be spawned. You can login through an Amazon
            account either in the spawned browser window, if available, or by opening http://your-ip:3000
            in the browser on another machine.

        :param audio_device: Name of the input audio device (default: 'default')
        :param audio_player: Player to be used for audio playback (default: 'default').
            Supported values: 'mpv', 'mpg123', 'gstreamer'
        """
        from avs.alexa import Alexa
        from avs.config import DEFAULT_CONFIG_FILE
        from avs.mic import Audio

        super().__init__(**kwargs)

        if not avs_config_file:
            avs_config_file = DEFAULT_CONFIG_FILE

        if not avs_config_file or not os.path.isfile(avs_config_file):
            from avs.auth import auth
            auth(None, avs_config_file)
            self.logger.warning('Amazon Echo assistant credentials not configured. Open http://localhost:3000 ' +
                                'to authenticate this client')

        self.audio_device = audio_device
        self.audio_player = audio_player
        self.audio = Audio(device_name=audio_device)
        self.alexa = Alexa(avs_config_file, audio_player=audio_player)
        self._ready = False

        self.alexa.state_listener.on_ready = self._on_ready()
        self.alexa.state_listener.on_listening = self._on_listening()
        self.alexa.state_listener.on_speaking = self._on_speaking()
        self.alexa.state_listener.on_thinking = self._on_thinking()
        self.alexa.state_listener.on_finished = self._on_finished()
        self.alexa.state_listener.on_disconnected = self._on_disconnected()

        self.audio.link(self.alexa)
        self.alexa.start()

    def _on_ready(self):
        def _callback():
            self._ready = True
        return _callback

    def _on_listening(self):
        def _callback():
            get_bus().post(ConversationStartEvent(assistant=self))
        return _callback

    def _on_speaking(self):
        def _callback():
            # AVS doesn't provide a way to access the response text
            get_bus().post(ResponseEvent(assistant=self, response_text=''))
        return _callback

    def _on_finished(self):
        def _callback():
            get_bus().post(ConversationEndEvent(assistant=self))
        return _callback

    def _on_disconnected(self):
        def _callback():
            self._ready = False
        return _callback

    def _on_thinking(self):
        def _callback():
            # AVS doesn't provide a way to access the detected text
            get_bus().post(SpeechRecognizedEvent(assistant=self, phrase=''))
        return _callback

    @action
    def start_conversation(self, **kwargs):
        if not self._ready:
            raise RuntimeError('Echo assistant not ready')

        self.audio.start()
        self.alexa.listen()

    @action
    def stop_conversation(self):
        self.audio.stop()
        self._on_finished()()


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class GooglePubsubMessageEvent(Event):
    """
    Event triggered when a new message is received on a subscribed Google Pub/Sub topic.
    """

    def __init__(self, topic: str, msg, *args, **kwargs):
        super().__init__(*args, topic=topic, msg=msg, *args, **kwargs)


# vim:sw=4:ts=4:et:



import json
import os
import time
import threading

from .. import Backend

from platypush.message import Message
from platypush.message.event import StopEvent
from platypush.message.request import Request
from platypush.message.response import Response


class LocalBackend(Backend):
    """
    Sends and receive messages on two distinct local FIFOs, one for
    the requests and one for the responses.

    You can use this backend either to send local commands to push through
    Pusher (or any other script), or debug. You can even send command on the
    command line and read the responses in this way::

        # Send the request. Remember: the JSON must be all on one line.
        $ cat <<EOF > /tmp/platypush-requests.fifo
        {"type": "request", "target": "node_name", "action": "shell.exec", "args": {"cmd":"echo ping"}, "origin": "node_name", "id": "put_an_id_here"}
        EOF
        $ cat /tmp/platypush-responses.fifo
        ping
        $
    """

    def __init__(self, request_fifo, response_fifo, **kwargs):
        super().__init__(**kwargs)

        self.request_fifo  = request_fifo
        self.response_fifo = response_fifo

        try: os.mkfifo(self.request_fifo)
        except FileExistsError as e: pass

        try: os.mkfifo(self.response_fifo)
        except FileExistsError as e: pass


    def send_message(self, msg):
        fifo = self.response_fifo \
            if isinstance(msg, Response) or self._request_context \
            else self.request_fifo

        msg = '{}\n'.format(str(msg)).encode('utf-8')

        with open(fifo, 'wb') as f:
            f.write(msg)


    def _get_next_message(self):
        fifo = self.response_fifo if self._request_context else self.request_fifo
        with open(fifo, 'rb', 0) as f:
            msg = f.readline()

        return Message.build(msg) if len(msg) else None


    def run(self):
        super().run()
        self.logger.info('Initialized local backend on {} and {}'.
                     format(self.request_fifo, self.response_fifo))

        while not self.should_stop():
            try:
                msg = self._get_next_message()
                if not msg: continue
            except Exception as e:
                self.logger.exception(e)
                time.sleep(0.2)
                continue

            self.logger.debug('Received message on the local backend: {}, thread_id: '.format(msg, self.thread_id))

            if self.should_stop(): break
            self.on_message(msg)


# vim:sw=4:ts=4:et:




import os
import sys

testdir = os.path.dirname(__file__)
sys.path.insert(0, os.path.abspath(os.path.join(testdir, '..')))
config_file = os.path.join(testdir, 'etc', 'config.yaml')

from platypush.config import Config
Config.init(config_file)

import platypush


class TestTimeoutException(RuntimeError):
    def __init__(self, msg):
        self.msg = msg


# vim:sw=4:ts=4:et:




from platypush.context import get_plugin
from platypush.plugins.media.search import MediaSearcher

class TorrentMediaSearcher(MediaSearcher):
    def search(self, query):
        self.logger.info('Searching torrents for "{}"'.format(query))

        torrents = get_plugin('torrent')
        if not torrents:
            raise RuntimeError('Torrent plugin not available/configured')
        return torrents.search(query, ).output



# vim:sw=4:ts=4:et:



import subprocess
import urllib.parse
from typing import Optional, List

from platypush.plugins import Plugin, action


class TtsPlugin(Plugin):
    """
    Default Text-to-Speech plugin. It leverages Google Translate.

    Requires:

        * **mplayer** - see your distribution docs on how to install the mplayer package
    """

    def __init__(self, language='en-gb', player_args: Optional[List[str]] = None):
        """
        :param language: Language code (default: ``en-gb``).
        :param player_args: Extra options to be passed to the audio player (default: ``mplayer``).
        """
        super().__init__()
        self.language = language
        self.player_args = player_args or []

    @action
    def say(self, text: str, language: Optional[str] = None, player_args: Optional[List[str]] = None):
        """
        Say some text.

        :param text: Text to say.
        :param language: Language code override.
        :param player_args: ``player_args`` override.
        """
        language = language or self.language
        player_args = player_args or self.player_args
        cmd = [
            'mplayer -ao alsa -really-quiet -noconsolecontrols ' +
            ' '.join(player_args) + ' ' +
            '"http://translate.google.com/translate_tts?{}"'.format(
                urllib.parse.urlencode({
                    'ie': 'UTF-8',
                    'client': 'tw-ob',
                    'tl': language,
                    'q': text,
                })
            )
        ]

        try:
            return subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, shell=True).decode('utf-8')
        except subprocess.CalledProcessError as e:
            raise RuntimeError(e.output.decode('utf-8'))

# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class MusicEvent(Event):
    """ Base class for music events """

    def __init__(self, status, track, *args, **kwargs):
        super().__init__(*args, status=status, track=track, **kwargs)


class MusicPlayEvent(MusicEvent):
    """
    Event fired upon music player playback start
    """

    def __init__(self, status=None, track=None, *args, **kwargs):
        """
        :param status: Player status
        :type status: dict

        :param track: Track being played
        :type track: dict
        """

        super().__init__(*args, status=status, track=track, **kwargs)


class MusicStopEvent(MusicEvent):
    """
    Event fired upon playback stop
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class MusicPauseEvent(MusicEvent):
    """
    Event fired upon playback paused
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SeekChangeEvent(MusicEvent):
    """
    Event fired upon seek change
    """

    def __init__(self, position, status=None, track=None, *args, **kwargs):
        super().__init__(position=position, status=status, track=track, *args, **kwargs)


class VolumeChangeEvent(MusicEvent):
    """
    Event fired upon volume change
    """

    def __init__(self, volume, status=None, track=None, *args, **kwargs):
        super().__init__(volume=volume, status=status, track=track, *args, **kwargs)


class MuteChangeEvent(MusicEvent):
    """
    Event fired upon mute change
    """

    def __init__(self, mute, status=None, track=None, *args, **kwargs):
        super().__init__(mute=mute, status=status, track=track, *args, **kwargs)


class PlaybackRepeatModeChangeEvent(MusicEvent):
    """
    Event fired upon repeat mode change
    """

    def __init__(self, state, status=None, track=None, *args, **kwargs):
        super().__init__(state=state, status=status, track=track, *args, **kwargs)


class PlaybackRandomModeChangeEvent(MusicEvent):
    """
    Event fired upon random mode change
    """

    def __init__(self, state, status=None, track=None, *args, **kwargs):
        super().__init__(state=state, status=status, track=track, *args, **kwargs)


class PlaybackConsumeModeChangeEvent(MusicEvent):
    """
    Event fired upon consume mode change
    """

    def __init__(self, state, status=None, track=None, *args, **kwargs):
        super().__init__(state=state, status=status, track=track, *args, **kwargs)


class PlaybackSingleModeChangeEvent(MusicEvent):
    """
    Event fired upon single mode change
    """

    def __init__(self, state, status=None, track=None, *args, **kwargs):
        super().__init__(state=state, status=status, track=track, *args, **kwargs)


class PlaylistChangeEvent(MusicEvent):
    """
    Event fired upon playlist change
    """

    def __init__(self, changes=None, status=None, track=None, *args, **kwargs):
        """
        :param changes: List with the tracks being added or removed
        :type changes: list

        :param status: Player status
        :type status: dict

        :param track: Track being played
        :type track: dict
        """

        super().__init__(changes=changes, status=status, track=track, *args, **kwargs)


class NewPlayingTrackEvent(MusicEvent):
    """
    Event fired when a new track is being played
    """

    def __init__(self, status=None, track=None, *args, **kwargs):
        """
        :param status: Player status
        :type status: dict

        :param track: Track being played
        :type track: dict
        """

        super().__init__(*args, status=status, track=track, **kwargs)


# vim:sw=4:ts=4:et:




from platypush.message.event import Event


class GoogleFitEvent(Event):
    """
    Event triggered upon new Google Fit data points
    """

    def __init__(self, data_source_id, values, *args, **kwargs):
        super().__init__(*args, data_source_id=data_source_id, values=values,
                         **kwargs)


# vim:sw=4:ts=4:et:



import os

from flask import abort, jsonify, request, send_from_directory, Blueprint

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.routes.plugins.media import media_map, \
    remove_subtitles, add_subtitles

subtitles = Blueprint('subtitles', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    subtitles,
]

@subtitles.route('/media/subtitles/<media_id>.vtt', methods=['GET', 'POST', 'DELETE'])
def handle_subtitles(media_id):
    """
    This route can be used to download and/or expose subtitle files
    associated to a media file
    """

    if request.method == 'GET':
        media_hndl = media_map.get(media_id)
        if not media_hndl:
            abort(404, 'No such media')

        if not media_hndl.subtitles:
            abort(404, 'The media has no subtitles attached')

        return send_from_directory(
            os.path.dirname(media_hndl.subtitles),
            os.path.basename(media_hndl.subtitles),
            mimetype='text/vtt')

    try:
        if request.method == 'DELETE':
            return jsonify(remove_subtitles(media_id))
        else:
            return jsonify(add_subtitles(media_id, request))
    except FileNotFoundError as e:
        abort(404, str(e))
    except AttributeError as e:
        abort(400, str(e))
    except NotImplementedError as e:
        abort(422, str(e))
    except Exception as e:
        abort(500, str(e))


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.message.event import Event


class TravisciBuildEvent(Event):
    def __init__(self,
                 repository_id: int,
                 repository_name: str,
                 repository_slug: str,
                 passed: bool,
                 build_id: int,
                 build_number: int,
                 duration: int,
                 previous_state: str,
                 private: bool,
                 tag: Optional[str],
                 branch: str,
                 commit_id: Optional[str],
                 commit_sha: Optional[str],
                 commit_message: Optional[str],
                 committed_at: str,
                 created_by: str,
                 started_at: str,
                 finished_at: str,
                 *args,
                 **kwargs):
        super().__init__(*args,
                         repository_id=repository_id,
                         repository_name=repository_name,
                         repository_slug=repository_slug,
                         passed=passed,
                         build_id=build_id,
                         build_number=build_number,
                         duration=duration,
                         previous_state=previous_state,
                         private=private,
                         tag=tag,
                         branch=branch,
                         commit_id=commit_id,
                         commit_sha=commit_sha,
                         commit_message=commit_message,
                         committed_at=committed_at,
                         created_by=created_by,
                         started_at=started_at,
                         finished_at=finished_at,
                         **kwargs)


class TravisciBuildPassedEvent(TravisciBuildEvent):
    """
    Event triggered when a Travis-Ci build passes.
    """
    def __init__(self, *args, **kwargs):
        kwargs['passed'] = True
        super().__init__(*args, **kwargs)


class TravisciBuildFailedEvent(TravisciBuildEvent):
    """
    Event triggered when a Travis-Ci build fails.
    """
    def __init__(self, *args, **kwargs):
        kwargs['passed'] = False
        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:



import datetime

from typing import Optional, Dict, List, Union

# noinspection PyPackageRequirements
import trello
# noinspection PyPackageRequirements
from trello.board import Board, List as List_
# noinspection PyPackageRequirements
from trello.exceptions import ResourceUnavailable

from platypush.message.response.trello import TrelloBoard, TrelloBoardsResponse, TrelloCardsResponse, TrelloCard, \
    TrelloAttachment, TrelloPreview, TrelloChecklist, TrelloChecklistItem, TrelloUser, TrelloComment, TrelloLabel, \
    TrelloList, TrelloBoardResponse, TrelloListsResponse, TrelloMembersResponse, TrelloMember, TrelloCardResponse

from platypush.plugins import Plugin, action


class TrelloPlugin(Plugin):
    """
    Trello integration.

    Requires:

        * **py-trello** (``pip install py-trello``)

    You'll also need a Trello API key. You can get it `here <https://trello.com/app-key>`.
    You'll also need an auth token if you want to view/change private resources. You can generate a permanent token
    linked to your account on
    https://trello.com/1/connect?key=<KEY>&name=platypush&response_type=token&expiration=never&scope=read,write
    """

    def __init__(self, api_key: str, api_secret: Optional[str] = None, token: Optional[str] = None, **kwargs):
        """
        :param api_key: Trello API key. You can get it `here <https://trello.com/app-key>`.
        :param api_secret: Trello API secret. You can get it `here <https://trello.com/app-key>`.
        :param token: Trello token. It is required if you want to access or modify private resources. You can get
            a permanent token on
            https://trello.com/1/connect?key=<KEY>&name=platypush&response_type=token&expiration=never&scope=read,write
        """

        super().__init__(**kwargs)
        self.api_key = api_key
        self.api_secret = api_secret
        self.token = token
        self._client = None

    def _get_client(self) -> trello.TrelloClient:
        if not self._client:
            self._client = trello.TrelloClient(
                api_key=self.api_key,
                api_secret=self.api_secret,
                token=self.token,
            )

        return self._client

    def _get_board(self, board: str) -> Board:
        client = self._get_client()
        try:
            return client.get_board(board)
        except ResourceUnavailable:
            boards = [b for b in client.list_boards() if b.name == board]
            assert boards, 'No such board: {}'.format(board)
            return boards[0]

    # noinspection PyShadowingBuiltins
    @action
    def get_boards(self, all: bool = False) -> TrelloBoardsResponse:
        """
        Get the list of boards.

        :param all: If True, return all the boards included those that have been closed/archived/deleted. Otherwise,
            only return open/active boards (default: False).
        """
        client = self._get_client()

        return TrelloBoardsResponse([
            TrelloBoard(
                id=b.id,
                name=b.name,
                description=b.description,
                url=b.url,
                date_last_activity=b.date_last_activity,
                closed=b.closed,
            )
            for b in client.list_boards(board_filter='all' if all else 'open')
        ])

    @action
    def get_board(self, board: str) -> TrelloBoardResponse:
        """
        Get the info about a board.

        :param board: Board ID or name
        """

        board = self._get_board(board)
        return TrelloBoardResponse(
            TrelloBoard(
                id=board.id,
                name=board.name,
                url=board.url,
                closed=board.closed,
                description=board.description,
                date_last_activity=board.date_last_activity,
                lists=[
                    TrelloList(id=ll.id, name=ll.name, closed=ll.closed, subscribed=ll.subscribed)
                    for ll in board.list_lists()
                ]
            )
        )

    @action
    def open_board(self, board: str):
        """
        Re-open/un-archive a board.

        :param board: Board ID or name
        """
        board = self._get_board(board)
        board.open()

    @action
    def close_board(self, board: str):
        """
        Close/archive a board.

        :param board: Board ID or name
        """
        board = self._get_board(board)
        board.close()

    @action
    def set_board_name(self, board: str, name: str):
        """
        Change the name of a board.

        :param board: Board ID or name.
        :param name: New name.
        """
        board = self._get_board(board)
        board.set_name(name)
        return self.get_board(name)

    @action
    def set_board_description(self, board: str, description: str):
        """
        Change the description of a board.

        :param board: Board ID or name.
        :param description: New description.
        """
        board = self._get_board(board)
        board.set_description(description)
        return self.get_board(description)

    @action
    def create_label(self, board: str, name: str, color: Optional[str] = None):
        """
        Add a label to a board.

        :param board: Board ID or name
        :param name: Label name
        :param color: Optional HTML color
        """
        board = self._get_board(board)
        board.add_label(name=name, color=color)

    @action
    def delete_label(self, board: str, label: str):
        """
        Delete a label from a board.

        :param board: Board ID or name
        :param label: Label ID or name
        """

        board = self._get_board(board)

        try:
            board.delete_label(label)
        except ResourceUnavailable:
            labels = [
                ll for ll in board.get_labels()
                if ll.name == label
            ]

            assert labels, 'No such label: {}'.format(label)
            label = labels[0].id
            board.delete_label(label)

    @action
    def add_member(self, board: str, member_id: str, member_type: str = 'normal'):
        """
        Add a member to a board.

        :param board: Board ID or name.
        :param member_id: Member ID to add.
        :param member_type: Member type - can be 'normal' or 'admin' (default: 'normal').
        """
        board = self._get_board(board)
        board.add_member(member_id, member_type=member_type)

    @action
    def remove_member(self, board: str, member_id: str):
        """
        Remove a member from a board.

        :param board: Board ID or name.
        :param member_id: Member ID to remove.
        """
        board = self._get_board(board)
        board.remove_member(member_id)

    def _get_members(self, board: str, only_admin: bool = False) -> TrelloMembersResponse:
        board = self._get_board(board)
        members = board.admin_members() if only_admin else board.get_members()

        return TrelloMembersResponse([
            TrelloMember(
                id=m.id,
                full_name=m.full_name,
                bio=m.bio,
                url=m.url,
                username=m.username,
                initials=m.initials,
                member_type=getattr(m, 'member_type') if hasattr(m, 'member_type') else None
            )
            for m in members
        ])

    @action
    def get_members(self, board: str) -> TrelloMembersResponse:
        """
        Get the list of all the members of a board.
        :param board: Board ID or name.
        """
        return self._get_members(board, only_admin=False)

    @action
    def get_admin_members(self, board: str) -> TrelloMembersResponse:
        """
        Get the list of the admin members of a board.
        :param board: Board ID or name.
        """
        return self._get_members(board, only_admin=True)

    # noinspection PyShadowingBuiltins
    @action
    def get_lists(self, board: str, all: bool = False) -> TrelloListsResponse:
        """
        Get the list of lists on a board.

        :param board: Board ID or name
        :param all: If True, return all the lists, included those that have been closed/archived/deleted. Otherwise,
            only return open/active lists (default: False).
        """

        board = self._get_board(board)
        return TrelloListsResponse([
            TrelloList(id=ll.id, name=ll.name, closed=ll.closed, subscribed=ll.subscribed)
            for ll in board.list_lists('all' if all else 'open')
        ])

    @action
    def add_list(self, board: str, name: str, pos: Optional[int] = None):
        """
        Add a list to a board.

        :param board: Board ID or name
        :param name: List name
        :param pos: Optional position (default: last)
        """

        board = self._get_board(board)
        board.add_list(name=name, pos=pos)

    # noinspection PyShadowingBuiltins
    def _get_list(self, board: str, list: str) -> List_:
        board = self._get_board(board)

        try:
            return board.get_list(list)
        except ResourceUnavailable:
            lists = [ll for ll in board.list_lists() if ll.name == list]
            assert lists, 'No such list: {}'.format(list)
            return lists[0]

    # noinspection PyShadowingBuiltins
    @action
    def set_list_name(self, board: str, list: str, name: str):
        """
        Change the name of a board list.

        :param board: Board ID or name
        :param list: List ID or name
        :param name: New name
        """
        list = self._get_list(board, list)
        list.set_name(name)

    # noinspection PyShadowingBuiltins
    @action
    def list_subscribe(self, board: str, list: str):
        """
        Subscribe to a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        list = self._get_list(board, list)
        list.subscribe()

    # noinspection PyShadowingBuiltins
    @action
    def list_unsubscribe(self, board: str, list: str):
        """
        Unsubscribe from a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        list = self._get_list(board, list)
        list.unsubscribe()

    # noinspection PyShadowingBuiltins
    @action
    def archive_all_cards(self, board: str, list: str):
        """
        Archive all the cards on a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        list = self._get_list(board, list)
        list.archive_all_cards()

    @action
    def move_all_cards(self, board: str, src: str, dest: str):
        """
        Move all the cards from a list to another.

        :param board: Board ID or name
        :param src: Source list
        :param dest: Target list
        """
        src = self._get_list(board, src)
        dest = self._get_list(board, dest)
        src.move_all_cards(dest.id)

    # noinspection PyShadowingBuiltins
    @action
    def open_list(self, board: str, list: str):
        """
        Open/un-archive a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        list = self._get_list(board, list)
        list.open()

    # noinspection PyShadowingBuiltins
    @action
    def close_list(self, board: str, list: str):
        """
        Close/archive a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        list = self._get_list(board, list)
        list.close()

    # noinspection PyShadowingBuiltins
    @action
    def move_list(self, board: str, list: str, position: int):
        """
        Move a list to another position.

        :param board: Board ID or name
        :param list: List ID or name
        :param position: New position index
        """
        list = self._get_list(board, list)
        list.move(position)

    # noinspection PyShadowingBuiltins
    @action
    def add_card(self, board: str, list: str, name: str, description: Optional[str] = None,
                 position: Optional[int] = None, labels: Optional[List[str]] = None,
                 due: Optional[Union[str, datetime.datetime]] = None, source: Optional[str] = None,
                 assign: Optional[List[str]] = None) -> TrelloCardResponse:
        """
        Add a card to a list.

        :param board: Board ID or name
        :param list: List ID or name
        :param name: Card name
        :param description: Card description
        :param position: Card position index
        :param labels: List of labels
        :param due: Due date (``datetime.datetime`` object or ISO-format string)
        :param source: Card ID to clone from
        :param assign: List of assignee member IDs
        """
        list = self._get_list(board, list)

        if labels:
            labels = [
                ll for ll in list.board.get_labels()
                if ll.id in labels or ll.name in labels
            ]

        card = list.add_card(name=name, desc=description, labels=labels, due=due, source=source, position=position,
                             assign=assign)

        return TrelloCardResponse(TrelloCard(id=card.id,
                                             name=card.name,
                                             url=card.url,
                                             closed=card.closed,
                                             board=TrelloBoard(
                                                 id=list.board.id,
                                                 name=list.board.name,
                                                 url=list.board.url,
                                                 closed=list.board.closed,
                                                 description=list.board.description,
                                                 date_last_activity=list.board.date_last_activity
                                             ),

                                             is_due_complete=card.is_due_complete,
                                             list=None,
                                             comments=[],
                                             labels=[
                                                 TrelloLabel(
                                                     id=lb.id,
                                                     name=lb.name,
                                                     color=lb.color
                                                 )
                                                 for lb in (card.labels or [])
                                             ],
                                             description=card.description,
                                             due_date=card.due_date,
                                             latest_card_move_date=card.latestCardMove_date,
                                             date_last_activity=card.date_last_activity
                                             ))

    @action
    def delete_card(self, card_id: str):
        """
        Permanently delete a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.delete()

    @action
    def open_card(self, card_id: str):
        """
        Open/un-archive a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_closed(False)

    @action
    def close_card(self, card_id: str):
        """
        Close/archive a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_closed(True)

    @action
    def add_checklist(self, card_id: str, title: str, items: List[str], states: Optional[List[bool]] = None):
        """
        Add a checklist to a card.

        :param card_id: Card ID
        :param title: Checklist title
        :param items: List of items in the checklist
        :param states: State of each item, True for checked, False for unchecked
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.add_checklist(title, items, states)

    @action
    def add_label(self, card_id: str, label: str):
        """
        Add a label to a card.

        :param card_id: Card ID
        :param label: Label name
        """
        client = self._get_client()
        card = client.get_card(card_id)

        labels = [
            ll for ll in card.board.get_labels()
            if ll.name == label
        ]

        assert labels, 'No such label: {}'.format(label)
        label = labels[0]
        card.add_label(label)

    @action
    def remove_label(self, card_id: str, label: str):
        """
        Remove a label from a card.

        :param card_id: Card ID
        :param label: Label name
        """
        client = self._get_client()
        card = client.get_card(card_id)

        labels = [
            ll for ll in card.board.get_labels()
            if ll.name == label
        ]

        assert labels, 'No such label: {}'.format(label)
        label = labels[0]
        card.remove_label(label)

    @action
    def assign_card(self, card_id: str, member_id: str):
        """
        Assign a card.

        :param card_id: Card ID
        :param member_id: Member ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.assign(member_id)

    @action
    def unassign_card(self, card_id: str, member_id: str):
        """
        Un-assign a card.

        :param card_id: Card ID
        :param member_id: Member ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.unassign(member_id)

    @action
    def attach_card(self, card_id: str, name: Optional[str] = None, mime_type: Optional[str] = None,
                    file: Optional[str] = None, url: Optional[str] = None):
        """
        Add an attachment to a card. It can be either a local file or a remote URL.

        :param card_id: Card ID
        :param name: File name
        :param mime_type: MIME type
        :param file: Path to the file
        :param url: URL to the file
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.attach(name=name, mimeType=mime_type, file=file, url=url)

    @action
    def remove_attachment(self, card_id: str, attachment_id: str):
        """
        Remove an attachment from a card.

        :param card_id: Card ID
        :param attachment_id: Attachment ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.remove_attachment(attachment_id)

    # noinspection PyShadowingBuiltins
    @action
    def change_card_board(self, card_id: str, board: str, list: str = None):
        """
        Move a card to a new board.

        :param card_id: Card ID
        :param board: New board ID or name
        :param list: Optional target list ID or name
        """
        client = self._get_client()
        card = client.get_card(card_id)
        board = self._get_board(board)

        list_id = None
        if list:
            list_id = self._get_list(board.id, list).id

        card.change_board(board.id, list_id)

    # noinspection PyShadowingBuiltins
    @action
    def change_card_list(self, card_id: str, list: str):
        """
        Move a card to a new list.

        :param card_id: Card ID
        :param list: List ID or name
        """
        client = self._get_client()
        card = client.get_card(card_id)
        list = self._get_list(card.board.id, list)
        card.change_list(list.id)

    @action
    def change_card_pos(self, card_id: str, position: int):
        """
        Move a card to a new position.

        :param card_id: Card ID
        :param position: New position index
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.change_pos(position)

    @action
    def comment_card(self, card_id: str, text: str):
        """
        Add a comment to a card.

        :param card_id: Card ID
        :param text: Comment text
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.comment(text)

    @action
    def update_comment(self, card_id: str, comment_id: str, text: str):
        """
        Update the content of a comment.

        :param card_id: Card ID
        :param comment_id: Comment ID
        :param text: New comment text
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.update_comment(comment_id, text)

    @action
    def delete_comment(self, card_id: str, comment_id: str):
        """
        Delete a comment.

        :param card_id: Card ID
        :param comment_id: Comment ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.delete_comment(comment_id)

    @action
    def set_card_name(self, card_id: str, name: str):
        """
        Change the name of a card.

        :param card_id: Card ID
        :param name: New name
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_name(name)

    @action
    def set_card_description(self, card_id: str, description: str):
        """
        Change the description of a card.

        :param card_id: Card ID
        :param description: New description
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_description(description)

    @action
    def add_card_member(self, card_id: str, member_id: str):
        """
        Add a member to a card.

        :param card_id: Card ID
        :param member_id: Member ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.add_member(member_id)

    @action
    def remove_card_member(self, card_id: str, member_id: str):
        """
        Remove a member from a card.

        :param card_id: Card ID
        :param member_id: Member ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.remove_member(member_id)

    @action
    def set_card_due(self, card_id: str, due: Union[str, datetime.datetime]):
        """
        Set the due date for a card.

        :param card_id: Card ID
        :param due: Due date, as a datetime.datetime object or an ISO string
        """
        client = self._get_client()
        card = client.get_card(card_id)

        if isinstance(due, str):
            due = datetime.datetime.fromisoformat(due)

        card.set_due(due)

    @action
    def remove_card_due(self, card_id: str):
        """
        Remove the due date from a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.remove_due()

    @action
    def set_card_due_complete(self, card_id: str):
        """
        Set the due date of a card as completed.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_due_complete()

    @action
    def remove_card_due_complete(self, card_id: str):
        """
        Remove the due complete flag from a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.remove_due_complete()

    @action
    def card_subscribe(self, card_id: str):
        """
        Subscribe to a card.
        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.subscribe()

    # noinspection PyShadowingBuiltins
    @action
    def get_cards(self, board: str, list: Optional[str] = None, all: bool = False) -> TrelloCardsResponse:
        """
        Get the list of cards on a board.

        :param board: Board ID or name
        :param list: List ID or name. If set then the method will only return the cards found on that list
            (default: None)
        :param all: If True, return all the cards included those that have been closed/archived/deleted. Otherwise,
            only return open/active cards (default: False).
        """

        board = self._get_board(board)
        lists: Dict[str, TrelloList] = {
            ll.id: TrelloList(id=ll.id, name=ll.name, closed=ll.closed, subscribed=ll.subscribed)
            for ll in board.list_lists()
        }

        list_id = None

        if list:
            if list in lists:
                list_id = list
            else:
                # noinspection PyUnresolvedReferences
                ll = [l1 for l1 in lists.values() if l1.name == list]
                assert ll, 'No such list ID/name: {}'.format(list)
                # noinspection PyUnresolvedReferences
                list_id = ll[0].id

        # noinspection PyUnresolvedReferences
        return TrelloCardsResponse([
            TrelloCard(
                id=c.id,
                name=c.name,
                url=c.url,
                closed=c.closed,
                list=lists.get(c.list_id),

                board=TrelloBoard(
                    id=c.board.id,
                    name=c.board.name,
                    url=c.board.url,
                    closed=c.board.closed,
                    description=c.board.description,
                    date_last_activity=c.board.date_last_activity
                ),

                attachments=[
                    TrelloAttachment(
                        id=a.get('id'),
                        bytes=a.get('bytes'),
                        date=a.get('date'),
                        edge_color=a.get('edgeColor'),
                        id_member=a.get('idMember'),
                        is_upload=a.get('isUpload'),
                        name=a.get('name'),
                        previews=[
                            TrelloPreview(
                                id=p.get('id'),
                                scaled=p.get('scaled'),
                                url=p.get('url'),
                                bytes=p.get('bytes'),
                                height=p.get('height'),
                                width=p.get('width')
                            )
                            for p in a.get('previews', [])
                        ],
                        url=a.get('url'),
                        mime_type=a.get('mimeType')
                    )
                    for a in c.attachments
                ],

                checklists=[
                    TrelloChecklist(
                        id=ch.id,
                        name=ch.name,
                        checklist_items=[
                            TrelloChecklistItem(
                                id=i.get('id'),
                                name=i.get('name'),
                                checked=i.get('checked')
                            )
                            for i in ch.items
                        ]
                    )
                    for ch in c.checklists
                ],

                comments=[
                    TrelloComment(
                        id=co.get('id'),
                        text=co.get('data', {}).get('text'),
                        type=co.get('type'),
                        date=co.get('date'),
                        creator=TrelloUser(
                            id=co.get('memberCreator', {}).get('id'),
                            username=co.get('memberCreator', {}).get('username'),
                            fullname=co.get('memberCreator', {}).get('fullName'),
                            initials=co.get('memberCreator', {}).get('initials'),
                            avatar_url=co.get('memberCreator', {}).get('avatarUrl')
                        )
                    )
                    for co in c.comments
                ],

                labels=[
                    TrelloLabel(
                        id=lb.id,
                        name=lb.name,
                        color=lb.color
                    )
                    for lb in (c.labels or [])
                ],

                is_due_complete=c.is_due_complete,
                due_date=c.due_date,
                description=c.description,
                latest_card_move_date=c.latestCardMove_date,
                date_last_activity=c.date_last_activity
            )
            for c in board.all_cards() if (
                (all or not c.closed) and
                (not list or c.list_id == list_id)
            )
        ])


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class WiimoteEvent(Event):
    """
    Event triggered upon Wiimote event
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class WiimoteConnectionEvent(WiimoteEvent):
    """
    Event triggered upon Wiimote connection
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class WiimoteDisconnectionEvent(WiimoteEvent):
    """
    Event triggered upon Wiimote disconnection
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:




from platypush.plugins import Plugin, action
from platypush.user import UserManager


class UserPlugin(Plugin):
    """
    Plugin to programmatically create and manage users and user sessions
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.user_manager = UserManager()

    @action
    def create_user(self, username, password, executing_user=None, executing_user_password=None, session_token=None,
                    **kwargs):
        """
        Create a user. This action needs to be executed by an already existing user, who needs to authenticate with
            their own credentials, unless this is the first user created on the system.

        :return: dict.

        Format::

            {
                "user_id": int,
                "username": str,
                "created_at": str (in ISO format)
            }

        """

        if self.user_manager.get_user_count() > 0 and not executing_user and not session_token:
            return None, "You need to authenticate in order to create another user"

        if not self.user_manager.authenticate_user(executing_user, executing_user_password):
            user, session = self.user_manager.authenticate_user_session(session_token)
            if not user:
                return None, "Invalid credentials and/or session_token"

        try:
            user = self.user_manager.create_user(username, password, **kwargs)
        except (NameError, ValueError) as e:
            return None, str(e)

        return {
            'user_id': user.user_id,
            'username': user.username,
            'created_at': user.created_at.isoformat(),
        }

    @action
    def authenticate_user(self, username, password):
        """
        Authenticate a user
        :return: True if the provided username and password are correct, False otherwise
        """

        return self.user_manager.authenticate_user(username, password)

    @action
    def update_password(self, username, old_password, new_password):
        """
        Update the password of a user
        :return: True if the password was successfully updated, false otherwise
        """

        return self.user_manager.update_password(username, old_password, new_password)

    @action
    def delete_user(self, username, executing_user=None, executing_user_password=None, session_token=None):
        """
        Delete a user
        """

        if not self.user_manager.authenticate_user(executing_user, executing_user_password):
            user, session = self.user_manager.authenticate_user_session(session_token)
            if not user:
                return None, "Invalid credentials and/or session_token"

        try:
            return self.user_manager.delete_user(username)
        except NameError:
            return None, "No such user: {}".format(username)

    @action
    def create_session(self, username, password, expires_at=None):
        """
        Create a user session
        :return: dict::

            {
                "session_token": str,
                "user_id": int,
                "created_at": str (in ISO format),
                "expires_at": str (in ISO format),
            }

        """

        session = self.user_manager.create_user_session(username=username,
                                                        password=password,
                                                        expires_at=expires_at)

        if not session:
            return None, "Invalid credentials"

        return {
            'session_token': session.session_token,
            'user_id': session.user_id,
            'created_at': session.created_at.isoformat(),
            'expires_at': session.expires_at.isoformat() if session.expires_at else None,
        }

    @action
    def authenticate_session(self, session_token):
        """
        Authenticate a session by token and return the associated user
        :return: dict.

        Format::

            {
                "user_id": int,
                "username": str,
                "created_at": str (in ISO format)
            }

        """

        user, session = self.user_manager.authenticate_user_session(session_token=session_token)
        if not user:
            return None, 'Invalid session token'

        return {
            'user_id': user.user_id,
            'username': user.username,
            'created_at': user.created_at.isoformat(),
        }

    @action
    def delete_session(self, session_token):
        """
        Delete a user session
        """

        return self.user_manager.delete_user_session(session_token)


# vim:sw=4:ts=4:et:



from platypush.context import get_plugin
from platypush.plugins import Plugin, action


class MediaPlexPlugin(Plugin):
    """
    Plugin to interact with a Plex media server

    Requires:

        * **plexapi** (``pip install plexapi``)
    """

    def __init__(self, server, username, password, *args, **kwargs):
        """
        :param server: Plex server name
        :type server: str

        :param username: Plex username
        :type username: str

        :param password: Plex password
        :type username: str
        """

        from plexapi.myplex import MyPlexAccount
        super().__init__(*args, **kwargs)

        self.resource = MyPlexAccount(username, password).resource(server)
        self._plex = None


    @property
    def plex(self):
        if not self._plex:
            self._plex = self.resource.connect()

        return self._plex


    @action
    def get_clients(self):
        """
        Get the list of active clients
        """

        return [{
            'device': c.device,
            'device_class': c.deviceClass,
            'local': c.local,
            'model': c.model,
            'platform': c.platform,
            'platform_version': c.platformVersion,
            'product': c.product,
            'state': c.state,
            'title': c.title,
            'version': c.version,
        } for c in self.plex.clients()]


    def _get_client(self, name):
        return self.plex.client(name)


    @action
    def search(self, section=None, title=None, **kwargs):
        """
        Return all the items matching the search criteria (default: all library items)

        :param section: Section to search (Movies, Shows etc.)
        :type section: str

        :param title: Full or partial title
        :type title: str

        :param kwargs: Search criteria - includes e.g. title, unwatched, director, genre etc.
        :type kwargs: dict
        """

        ret = []
        library = self.plex.library

        if section:
            library = library.section(section)

        if title or kwargs:
            items = library.search(title, **kwargs)
        else:
            items = library.all()

        for item in items:
            ret.append(self._flatten_item(item))

        return ret


    @action
    def playlists(self):
        """
        Get the playlists on the server
        """

        return [
            {
                'title': pl.title,
                'duration': pl.duration,
                'summary': pl.summary,
                'viewed_at': pl.viewedAt,
                'items': [ self._flatten_item(item) for item in pl.items() ],
            } for pl in self.plex.playlists()
        ]


    @action
    def history(self):
        """
        Get the history of items played on the server
        """

        return [
            self._flatten_item(item) for item in self.plex.history()
        ]


    def get_chromecast(self, chromecast):
        from .lib.plexcast import PlexController

        hndl = PlexController()
        hndl.namespace = 'urn:x-cast:com.google.cast.sse'
        cast = get_plugin('media.chromecast').get_chromecast(chromecast)
        cast.register_handler(hndl)

        return (cast, hndl)


    @action
    def play(self, client=None, chromecast=None, **kwargs):
        """
        Search and play content on a client or a Chromecast. If no search filter
        is specified, a play event will be sent to the specified client.

        NOTE: Adding and managing play queues through the Plex API isn't fully
        supported yet, therefore in case multiple items are returned from the
        search only the first one will be played.

        :param client: Client name
        :type client: str

        :param chromecast: Chromecast name
        :type chromecast: str

        :param kwargs: Search filter (e.g. title, section, unwatched, director etc.)
        :type kwargs: dict
        """

        if not client and not chromecast:
            raise RuntimeError('No client nor chromecast specified')

        if client:
            client = plex.client(client)
        elif chromecast:
            (chromecast, handler) = self.get_chromecast(chromecast)

        if not kwargs:
            if client:
                return client.play()
            elif chromecast:
                return handler.play()

        if 'section' in kwargs:
            library = self.plex.library.section(kwargs.pop('section'))
        else:
            library = self.plex.library

        results = library.search(**kwargs)
        if not results:
            self.logger.info('No results for {}'.format(kwargs))
            return

        item = results[0]
        self.logger.info('Playing {} on {}'.format(item.title, client or chromecast))

        if client:
            return client.playMedia(item)
        elif chromecast:
            return handler.play_media(item, self.plex)


    @action
    def pause(self, client):
        """
        Send pause event to a client
        """

        return self.client(client).pause()


    @action
    def stop(self, client):
        """
        Send stop event to a client
        """

        return self.client(client).stop()


    @action
    def seek(self, client, offset):
        """
        Send seek event to a client
        """

        return self.client(client).seekTo(offset)


    @action
    def forward(self, client):
        """
        Forward playback on a client
        """

        return self.client(client).stepForward()


    @action
    def back(self, client):
        """
        Backward playback on a client
        """

        return self.client(client).stepBack()


    @action
    def next(self, client):
        """
        Play next item on a client
        """

        return self.client(client).skipNext()


    @action
    def previous(self, client):
        """
        Play previous item on a client
        """

        return self.client(client).skipPrevious()


    @action
    def set_volume(self, client, volume):
        """
        Set the volume on a client between 0 and 100
        """

        return self.client(client).setVolume(volume/100)


    @action
    def repeat(self, client, repeat):
        """
        Set the repeat status on a client
        """

        return self.client(client).setRepeat(repeat)


    @action
    def random(self, client, random):
        """
        Set the random status on a client
        """

        return self.client(client).setShuffle(random)


    @action
    def up(self, client):
        """
        Send an up key event to a client
        """

        return self.client(client).moveUp()


    @action
    def down(self, client):
        """
        Send a down key event to a client
        """

        return self.client(client).moveDown()


    @action
    def left(self, client):
        """
        Send a left key event to a client
        """

        return self.client(client).moveLeft()


    @action
    def right(self, client):
        """
        Send a right key event to a client
        """

        return self.client(client).moveRight()


    @action
    def go_back(self, client):
        """
        Send a back key event to a client
        """

        return self.client(client).goBack()


    @action
    def go_home(self, client):
        """
        Send a home key event to a client
        """

        return self.client(client).goHome()


    @action
    def go_to_media(self, client):
        """
        Send a go to media event to a client
        """

        return self.client(client).goToMedia()


    @action
    def go_to_music(self, client):
        """
        Send a go to music event to a client
        """

        return self.client(client).goToMusic()


    @action
    def next_letter(self, client):
        """
        Send a next letter event to a client
        """

        return self.client(client).nextLetter()


    @action
    def page_down(self, client):
        """
        Send a page down event to a client
        """

        return self.client(client).pageDown()


    @action
    def page_up(self, client):
        """
        Send a page up event to a client
        """

        return self.client(client).pageUp()


    def _flatten_item(self, item):
        from plexapi.video import Movie, Show

        _item = {
            'summary': item.summary,
            'title': item.title,
            'type': item.type,
            'genres': [g.tag for g in getattr(item, 'genres', [])],
            'art': getattr(item, 'art', None),
            'art_url': getattr(item, 'artUrl', None),
            'rating': getattr(item, 'rating', None),
            'content_rating': getattr(item, 'content_rating', None),
        }

        if isinstance(item, Movie):
            _item['is_watched'] = item.isWatched
            _item['view_offset'] = item.viewOffset
            _item['view_count'] = item.viewCount
            _item['year'] = item.year
            _item['audience_rating'] = item.audienceRating
            _item['countries'] = [c.tag for c in item.countries]

            _item['media'] = [
                {
                    'duration': (item.media[i].duration or 0)/1000,
                    'width': item.media[i].width,
                    'height': item.media[i].height,
                    'audio_channels': item.media[i].audioChannels,
                    'audio_codec': item.media[i].audioCodec,
                    'video_codec': item.media[i].videoCodec,
                    'video_resolution': item.media[i].videoResolution,
                    'video_frame_rate': item.media[i].videoFrameRate,
                    'parts': [
                        {
                            'file': part.file,
                            'duration': (part.duration or 0)/1000,
                        } for part in item.media[i].parts
                    ]
                } for i in range(0, len(item.media))
            ]

        elif isinstance(item, Show):
            _item['media'] = [
                {
                    'title': season.title,
                    'season_number': season.seasonNumber,
                    'summary': season.summary,
                    'episodes': [
                        {
                            'duration': episode.duration/1000,
                            'index': episode.index,
                            'year': episode.year,
                            'season_number': episode.seasonNumber,
                            'season_episode': episode.seasonEpisode,
                            'summary': episode.summary,
                            'is_watched': episode.isWatched,
                            'view_count': episode.viewCount,
                            'view_offset': episode.viewOffset,
                            'media': [
                                {
                                    'duration': episode.media[i].duration/1000,
                                    'width': episode.media[i].width,
                                    'height': episode.media[i].height,
                                    'audio_channels': episode.media[i].audioChannels,
                                    'audio_codec': episode.media[i].audioCodec,
                                    'video_codec': episode.media[i].videoCodec,
                                    'video_resolution': episode.media[i].videoResolution,
                                    'video_frame_rate': episode.media[i].videoFrameRate,
                                    'title': episode.title,
                                    'parts': [
                                        {
                                            'file': part.file,
                                            'duration': part.duration/1000,
                                        } for part in episode.media[i].parts
                                    ]
                                } for i in range(0, len(episode.locations))
                            ]
                        } for episode in season.episodes()
                    ]
                } for season in item.seasons()
            ]

        return _item


# vim:sw=4:ts=4:et:




from platypush.backend.sensor import SensorBackend


class SensorDistanceVl53L1XBackend(SensorBackend):
    """
    Backend to poll an `VL53L1x <https://www.st.com/en/imaging-and-photonics-solutions/vl53l1x.html>`_
    laser ranger/distance sensor

    Requires:

        * ``smbus2`` (``pip install smbus2``)
        * ``vl53l1x`` (``pip install vl53l1x``)
    """

    def __init__(self, short=True, medium=False, long=False, **kwargs):
        """
        :param short: Enable short range measurement (default: True)
        :param medium: Enable medium range measurement (default: False)
        :param long: Enable long range measurement (default: False)
        """
        enabled_sensors = {
            'short': short,
            'medium': medium,
            'long': long,
        }

        super().__init__(plugin='gpio.sensor.distance.vl53l1x', enabled_sensors=enabled_sensors, **kwargs)


# vim:sw=4:ts=4:et:



import io
import os
from typing import Optional, List, Union

from platypush.plugins import action
from platypush.plugins.google import GooglePlugin
from platypush.message.response.google.drive import GoogleDriveFile


class GoogleDrivePlugin(GooglePlugin):
    """
    Google Drive plugin.
    """

    scopes = ['https://www.googleapis.com/auth/drive',
              'https://www.googleapis.com/auth/drive.appfolder',
              'https://www.googleapis.com/auth/drive.photos.readonly']

    def __init__(self, *args, **kwargs):
        super().__init__(scopes=self.scopes, *args, **kwargs)

    def get_service(self, **kwargs):
        return super().get_service(service='drive', version='v3')

    # noinspection PyShadowingBuiltins
    @action
    def files(self,
              filter: Optional[str] = None,
              folder_id: Optional[str] = None,
              limit: Optional[int] = 100,
              drive_id: Optional[str] = None,
              spaces: Optional[Union[str, List[str]]] = None,
              order_by: Optional[Union[str, List[str]]] = None) -> Union[GoogleDriveFile, List[GoogleDriveFile]]:
        """
        Get the list of files.

        :param filter: Optional filter (default: None). See
            `Google Drive API docs <https://developers.google.com/drive/api/v3/search-files>`_ for
            the supported syntax.

        :param folder_id: Drive folder ID to search (default: get all files).
        :param limit: Maximum number of entries to be retrieves (default: 100).
        :param drive_id: Shared drive ID to search (default: None).
        :param spaces: Drive spaces to search. Supported values:

            - ``drive``
            - ``appDataFolder``
            - ``photos``

        :param order_by: Order the results by a specific attribute or list
            of attributes (default: None). Supported attributes:

                - ``createdTime``
                - ``folder``
                - ``modifiedByMeTime``
                - ``modifiedTime``
                - ``name``
                - ``name_natural``
                - ``quotaBytesUsed``
                - ``recency``
                - ``sharedWithMeTime``
                - ``starred``
                - ``viewedByMeTime``

            Attributes will be sorted in ascending order by default. You can change that by
            by appending "``desc``" separated by a space to the attribute you want in descending
            order - e.g. ``["folder", "createdTime desc", "modifiedTime desc"]``.
        """

        service = self.get_service()
        page_token = None
        files = []

        if isinstance(order_by, list):
            order_by = ','.join(order_by)
        if isinstance(spaces, list):
            spaces = ','.join(spaces)
        if folder_id:
            if not filter:
                filter = ''
            else:
                filter += ' '

            filter += "'{}' in parents".format(folder_id)

        while True:
            results = service.files().list(
                q=filter,
                driveId=drive_id,
                pageSize=limit,
                orderBy=order_by,
                fields="nextPageToken, files(id, name, kind, mimeType)",
                pageToken=page_token,
                spaces=spaces,
            ).execute()

            page_token = results.get('nextPageToken')
            files.extend([
                GoogleDriveFile(
                    id=f.get('id'),
                    name=f.get('name'),
                    type=f.get('kind').split('#')[1],
                    mime_type=f.get('mimeType'),
                ) for f in results.get('files', [])
            ])

            if not page_token or (limit and len(files) >= limit):
                break

        return files

    @action
    def get(self, file_id: str):
        """
        Get the information of a file on the Drive by file ID.
        :param file_id: File ID.
        """
        service = self.get_service()
        file = service.files().get(fileId=file_id).execute()
        return GoogleDriveFile(
            type=file.get('kind').split('#')[1],
            id=file.get('id'),
            name=file.get('name'),
            mime_type=file.get('mimeType'),
        )

    @action
    def upload(self,
               path: str,
               mime_type: Optional[str] = None,
               name: Optional[str] = None,
               description: Optional[str] = None,
               parents: Optional[List[str]] = None,
               starred: bool = False,
               target_mime_type: Optional[str] = None) -> GoogleDriveFile:
        """
        Upload a file to Google Drive.

        :param path: Path of the file to upload.
        :param mime_type: MIME type of the source file (e.g. "``image/jpeg``").
        :param name: Name of the target file. Default: same name as the source file.
        :param description: File description.
        :param parents: List of folder IDs that will contain the file (default: drive root).
        :param starred: If True, then the uploaded file will be marked as starred by the user.
        :param target_mime_type: Target MIME type. Useful if you want to e.g. import a CSV file as a Google Sheet
            (use "``application/vnd.google-apps.spreadsheet``), or an ODT file to a Google Doc
            (use "``application/vnd.google-apps.document``). See
            `the official documentation <https://developers.google.com/drive/api/v3/mime-types>`_ for a complete list
            of supported types.
        """
        # noinspection PyPackageRequirements
        from googleapiclient.http import MediaFileUpload

        path = os.path.abspath(os.path.expanduser(path))
        name = name or os.path.basename(path)
        metadata = {
            'name': name,
            'description': description,
            'parents': parents,
            'starred': starred,
        }

        if target_mime_type:
            metadata['mimeType'] = target_mime_type

        media = MediaFileUpload(path, mimetype=mime_type)
        service = self.get_service()
        file = service.files().create(
            body=metadata,
            media_body=media,
            fields='*'
        ).execute()

        return GoogleDriveFile(
            type=file.get('kind').split('#')[1],
            id=file.get('id'),
            name=file.get('name'),
            mime_type=file.get('mimeType'),
        )

    @action
    def download(self, file_id: str, path: str) -> str:
        """
        Download a Google Drive file locally.

        :param file_id: Path of the file to upload.
        :param path: Path of the file to upload.
        :return: The local file path.
        """
        # noinspection PyPackageRequirements
        from googleapiclient.http import MediaIoBaseDownload

        service = self.get_service()
        request = service.files().get_media(fileId=file_id)
        path = os.path.abspath(os.path.expanduser(path))
        if os.path.isdir(path):
            name = service.files().get(fileId=file_id).execute().get('name')
            path = os.path.join(path, name)

        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False

        while not done:
            status, done = downloader.next_chunk()
            self.logger.info('Download progress: {}%'.format(status.progress()))

        with open(path, 'wb') as f:
            f.write(fh.getbuffer().tobytes())
        return path

    @action
    def create(self,
               name: str,
               description: Optional[str] = None,
               mime_type: Optional[str] = None,
               parents: Optional[List[str]] = None,
               starred: bool = False) -> GoogleDriveFile:
        """
        Create a file.

        :param name: File name.
        :param description: File description.
        :param mime_type: File MIME type.
        :param parents: List of folder IDs that will contain the file (default: drive root).
        :param starred: If True then the file will be marked as starred.
        """
        metadata = {
            'name': name,
            'description': description,
            'parents': parents,
            'starred': starred,
        }

        if mime_type:
            metadata['mimeType'] = mime_type

        service = self.get_service()
        file = service.files().create(
            body=metadata,
            fields='*'
        ).execute()

        return GoogleDriveFile(
            type=file.get('kind').split('#')[1],
            id=file.get('id'),
            name=file.get('name'),
            mime_type=file.get('mimeType'),
        )

    @action
    def update(self,
               file_id: str,
               name: Optional[str] = None,
               description: Optional[str] = None,
               add_parents: Optional[List[str]] = None,
               remove_parents: Optional[List[str]] = None,
               mime_type: Optional[str] = None,
               starred: bool = None,
               trashed: bool = None) -> GoogleDriveFile:
        """
        Update the metadata or the content of a file.

        :param file_id: File ID.
        :param name: Set the file name.
        :param description: Set the file description.
        :param add_parents: Add the file to these parent folder IDs.
        :param remove_parents: Remove the file from these parent folder IDs.
        :param mime_type: Set the file MIME type.
        :param starred: Change the starred flag.
        :param trashed: Move/remove from trash.
        """
        metadata = {}
        if name:
            metadata['name'] = name
        if description is not None:
            metadata['description'] = description
        if add_parents:
            metadata['add_parents'] = add_parents
        if remove_parents:
            metadata['remove_parents'] = remove_parents
        if mime_type:
            metadata['mimeType'] = mime_type
        if starred is not None:
            metadata['starred'] = starred
        if trashed is not None:
            metadata['trashed'] = trashed

        service = self.get_service()
        file = service.files().update(
            fileId=file_id,
            body=metadata,
            fields='*'
        ).execute()

        return GoogleDriveFile(
            type=file.get('kind').split('#')[1],
            id=file.get('id'),
            name=file.get('name'),
            mime_type=file.get('mimeType'),
        )

    @action
    def delete(self, file_id: str):
        """
        Delete a file from Google Drive.
        :param file_id: File ID.
        """
        service = self.get_service()
        service.files().delete(fileId=file_id).execute()

    @action
    def copy(self, file_id: str) -> GoogleDriveFile:
        """
        Create a copy of a file.
        :param file_id: File ID.
        """
        service = self.get_service()
        file = service.files().copy(fileId=file_id).execute()
        return GoogleDriveFile(
            type=file.get('kind').split('#')[1],
            id=file.get('id'),
            name=file.get('name'),
            mime_type=file.get('mimeType'),
        )

    @action
    def empty_trash(self):
        """
        Empty the Drive bin.
        """
        service = self.get_service()
        service.files().emptyTrash()


# vim:sw=4:ts=4:et:



import copy
import importlib
import json
import logging
import re
import requests
import time

from datetime import date
from frozendict import frozendict
from threading import Thread

from platypush.message.event.http import HttpEvent
from platypush.utils import set_thread_name

class HttpRequest(object):
    poll_seconds = 60
    timeout = 5


    class HttpRequestArguments(object):
        def __init__(self, url, method='get', *args, **kwargs):
            self.method = method.lower()
            self.url = url
            self.args = args
            self.kwargs = kwargs


    def __init__(self, args, bus=None, poll_seconds=None, timeout=None,
                 skip_first_call=True, **kwargs):
        super().__init__()

        self.poll_seconds = poll_seconds or self.poll_seconds
        self.timeout = timeout or self.timeout
        self.bus = bus
        self.skip_first_call = skip_first_call
        self.last_request_timestamp = 0
        self.logger = logging.getLogger(__name__)

        if isinstance(args, self.HttpRequestArguments):
            self.args = args
        elif isinstance(args, dict):
            self.args = self.HttpRequestArguments(**args)
        else:
            raise RuntimeError('{} is neither a dictionary nor an HttpRequest')

        if 'timeout' not in self.args.kwargs:
            self.args.kwargs['timeout'] = self.timeout

        self.request_args = {
            'method': self.args.method, 'url': self.args.url, **self.args.kwargs
        }


    def execute(self):
        def _thread_func():
            set_thread_name('HttpPoll')
            is_first_call = self.last_request_timestamp == 0
            self.last_request_timestamp = time.time()

            try:
                method = getattr(requests, self.args.method.lower())
                response = method(self.args.url, *self.args.args, **self.args.kwargs)
                new_items = self.get_new_items(response)

                if isinstance(new_items, HttpEvent):
                    event = new_items
                    new_items = event.args['response']
                else:
                    event = HttpEvent(dict(self), new_items)

                if new_items and self.bus:
                    if not self.skip_first_call or (
                            self.skip_first_call and not is_first_call):
                        self.bus.post(event)

                response.raise_for_status()
            except Exception as e:
                self.logger.warning('Encountered an error while retrieving {}: {}'.
                                    format(self.args.url, str(e)))

        Thread(target=_thread_func, name='HttpPoll').start()


    def get_new_items(self, response):
        """ Gets new items out of a response """
        raise("get_new_items must be implemented in a derived class")


    def __iter__(self):
        for (key, value) in self.request_args.items():
            yield (key, value)


class JsonHttpRequest(HttpRequest):
    def __init__(self, path=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = path
        self.seen_entries = set()


    def get_new_items(self, response):
        response = response.json()
        new_entries = []

        if self.path:
            m = re.match('\$\{\s*(.*)\s*\}', self.path)
            response = eval(m.group(1))

        for entry in response:
            flattened_entry = deep_freeze(entry)
            if flattened_entry not in self.seen_entries:
                new_entries.append(entry)
            self.seen_entries.add(flattened_entry)

        return new_entries


def deep_freeze(x):
    if isinstance(x, str) or not hasattr(x, "__len__") :
        return x
    if hasattr(x, "keys") and hasattr(x, "values") :
        return frozendict({deep_freeze(k) : deep_freeze(v) for k,v in x.items()})
    if hasattr(x, "__getitem__") :
        return tuple(map(deep_freeze, x))

    return frozenset(map(deep_freeze,x))


# vim:sw=4:ts=4:et:




"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import time

from sqlalchemy import create_engine, Table, MetaData
from sqlalchemy.engine import Engine

from platypush.plugins import Plugin, action


class DbPlugin(Plugin):
    """
    Database plugin. It allows you to programmatically select, insert, update
    and delete records on a database backend through requests, procedures and
    event hooks.

    Requires:
        * **sqlalchemy** (``pip install sqlalchemy``)
    """

    engine = None
    _db_error_wait_interval = 5.0
    _db_error_retries = 3

    def __init__(self, engine=None, *args, **kwargs):
        """
        :param engine: Default SQLAlchemy connection engine string (e.g.  ``sqlite:///:memory:`` or ``mysql://user:pass@localhost/test``) that will be used. You can override the default engine in the db actions.
        :type engine: str
        :param args: Extra arguments that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        :param kwargs: Extra kwargs that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        """

        super().__init__()
        self.engine = self._get_engine(engine, *args, **kwargs)

    def _get_engine(self, engine=None, *args, **kwargs):
        if engine:
            if isinstance(engine, Engine):
                return engine

            return create_engine(engine, *args, **kwargs)

        return self.engine

    @staticmethod
    def _build_condition(table, column, value):
        if isinstance(value, str):
            value = "'{}'".format(value)
        elif not isinstance(value, int) and not isinstance(value, float):
            value = "'{}'".format(str(value))

        return eval('table.c.{}=={}'.format(column, value))

    @action
    def execute(self, statement, engine=None, *args, **kwargs):
        """
        Executes a raw SQL statement.

        .. warning::
            Avoid calling this method directly if possible.  Use ``insert``,
            ``update`` and ``delete`` methods instead if possible.  Don't use this
            method if you need to select records, use the ``select`` method
            instead, as this method is mostly meant to execute raw SQL without
            returning anything.

        :param statement: SQL to be executed
        :type statement: str
        :param engine: Engine to be used (default: default class engine)
        :type engine: str
        :param args: Extra arguments that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        :param kwargs: Extra kwargs that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        """

        engine = self._get_engine(engine, *args, **kwargs)

        with engine.connect() as connection:
            result = connection.execute(statement)

    def _get_table(self, table, engine=None, *args, **kwargs):
        if not engine:
            engine = self._get_engine(engine, *args, **kwargs)

        db_ok = False
        n_tries = 0
        last_error = None

        while not db_ok and n_tries < self._db_error_retries:
            try:
                n_tries += 1
                metadata = MetaData()
                table = Table(table, metadata, autoload=True, autoload_with=engine)
                db_ok = True
            except Exception as e:
                last_error = e
                wait_time = self._db_error_wait_interval * n_tries
                self.logger.exception(e)
                self.logger.info('Waiting {} seconds before retrying'.format(wait_time))
                time.sleep(wait_time)
                engine = self._get_engine(engine, *args, **kwargs)

        if not db_ok and last_error:
            raise last_error

        return table, engine


    @action
    def select(self, query=None, table=None, filter=None, engine=None, *args, **kwargs):
        """
        Returns rows (as a list of hashes) given a query.

        :param query: SQL to be executed
        :type query: str
        :param filter: Query WHERE filter expressed as a dictionary. This approach is preferred over specifying raw SQL
            in ``query`` as the latter approach may be prone to SQL injection, unless you need to build some complex
            SQL logic.
        :type filter: dict
        :param table: If you specified a filter instead of a raw query, you'll have to specify the target table
        :type table: str
        :param engine: Engine to be used (default: default class engine)
        :type engine: str
        :param args: Extra arguments that will be passed to ``sqlalchemy.create_engine``
            (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        :param kwargs: Extra kwargs that will be passed to ``sqlalchemy.create_engine``
            (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        :returns: List of hashes representing the result rows.

        Examples:

            Request::

                {
                    "type": "request",
                    "target": "your_host",
                    "action": "db.select",
                    "args": {
                        "engine": "sqlite:///:memory:",
                        "query": "SELECT id, name FROM table"
                    }
                }

            or::

                {
                    "type": "request",
                    "target": "your_host",
                    "action": "db.select",
                    "args": {
                        "engine": "sqlite:///:memory:",
                        "table": "table",
                        "filter": {"id": 1}
                    }
                }

            Response::

                [
                    {
                        "id": 1,
                        "name": foo
                    }
                ]
        """

        engine = self._get_engine(engine, *args, **kwargs)

        if table:
            table, engine = self._get_table(table, engine=engine, *args, **kwargs)
            query = table.select()

            if filter:
                for (k,v) in filter.items():
                    query = query.where(self._build_condition(table, k, v))

        if query is None:
            raise RuntimeError('You need to specify either "query", or "table" and "filter"')

        with engine.connect() as connection:
            result = connection.execute(query)

            columns = result.keys()
            rows = [
                { columns[i]: row[i] for i in range(0, len(columns)) }
                for row in result.fetchall()
            ]

        return rows

    @action
    def insert(self, table, records, engine=None, key_columns=None,
               on_duplicate_update=False, *args, **kwargs):
        """
        Inserts records (as a list of hashes) into a table.

        :param table: Table name
        :type table: str
        :param records: Records to be inserted (as a list of hashes)
        :type records: list
        :param engine: Engine to be used (default: default class engine)
        :type engine: str
        :param key_columns: Set it to specify the names of the key columns for ``table``. Set it if you want your statement to be executed with the ``on_duplicate_update`` flag.
        :type key_columns: list
        :param on_duplicate_update: If set, update the records in case of duplicate rows (default: False). If set, you'll need to specify ``key_columns`` as well.
        :type on_duplicate_update: bool
        :param args: Extra arguments that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        :param kwargs: Extra kwargs that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)

        Example:

            Request::

                {
                    "type": "request",
                    "target": "your_host",
                    "action": "db.insert",
                    "args": {
                        "table": "table",
                        "engine": "sqlite:///:memory:",
                        "records": [
                            {
                                "id": 1,
                                "name": foo
                            },

                            {
                                "id": 2,
                                "name": bar
                            }
                        ]
                    }
                }
        """

        if key_columns is None:
            key_columns = []

        engine = self._get_engine(engine, *args, **kwargs)

        for record in records:
            table, engine = self._get_table(table, engine=engine, *args, **kwargs)
            insert = table.insert().values(**record)

            try:
                engine.execute(insert)
            except Exception as e:
                if on_duplicate_update and key_columns:
                    self.update(table=table, records=records,
                                key_columns=key_columns, engine=engine,
                                *args, **kwargs)
                else:
                    raise e

    @action
    def update(self, table, records, key_columns, engine=None, *args, **kwargs):
        """
        Updates records on a table.

        :param table: Table name
        :type table: str
        :param records: Records to be updated (as a list of hashes)
        :type records: list
        :param key_columns: Names of the key columns, used in the WHERE condition
        :type key_columns: list
        :param engine: Engine to be used (default: default class engine)
        :type engine: str
        :param args: Extra arguments that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        :param kwargs: Extra kwargs that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)

        Example:

            Request::

                {
                    "type": "request",
                    "target": "your_host",
                    "action": "db.update",
                    "args": {
                        "table": "table",
                        "engine": "sqlite:///:memory:",
                        "key_columns": ["id"],
                        "records": [
                            {
                                "id": 1,
                                "name": foo
                            },

                            {
                                "id": 2,
                                "name": bar
                            }
                        ]
                    }
                }
        """

        engine = self._get_engine(engine, *args, **kwargs)

        for record in records:
            table, engine = self._get_table(table, engine=engine, *args, **kwargs)
            key = { k:v for (k,v) in record.items() if k in key_columns }
            values = { k:v for (k,v) in record.items() if k not in key_columns }

            update = table.update()

            for (k,v) in key.items():
                update = update.where(self._build_condition(table, k, v))

            update = update.values(**values)
            engine.execute(update)

    @action
    def delete(self, table, records, engine=None, *args, **kwargs):
        """
        Deletes records from a table.

        :param table: Table name
        :type table: str
        :param records: Records to be deleted, as a list of dictionaries
        :type records: list
        :param engine: Engine to be used (default: default class engine)
        :type engine: str
        :param args: Extra arguments that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)
        :param kwargs: Extra kwargs that will be passed to ``sqlalchemy.create_engine`` (see http://docs.sqlalchemy.org/en/latest/core/engines.html)

        Example:

            Request::

                {
                    "type": "request",
                    "target": "your_host",
                    "action": "db.delete",
                    "args": {
                        "table": "table",
                        "engine": "sqlite:///:memory:",
                        "records": [
                            { "id": 1 },
                            { "id": 2 }
                        ]
                    }
                }
        """

        engine = self._get_engine(engine, *args, **kwargs)

        for record in records:
            table, engine = self._get_table(table, engine=engine, *args, **kwargs)
            delete = table.delete()

            for (k,v) in record.items():
                delete = delete.where(self._build_condition(table, k, v))

            engine.execute(delete)


# vim:sw=4:ts=4:et:



import datetime
import dateutil.parser
import json

from platypush.backend.http.request import JsonHttpRequest
from platypush.message.event.http.ota.booking import NewReservationEvent


class GetReservationUpdates(JsonHttpRequest):
    """ Gets the reservation updates """
    def __init__(self, hotel_id, token, *args, **kwargs):
        self.hotel_id = hotel_id
        self.token = token
        self.seen_entries = set()
        self.last_update = None

        args = {
            'method': 'get',
            'url': 'https://hub-api.booking.com/v1/hotels/{}/reservations'.format(self.hotel_id),
            'headers': { 'X-Booking-Auth-Token': self.token },
            'params': { 'updatedSince': datetime.date.today().isoformat() }
        }

        super().__init__(args=args, **kwargs)


    def get_new_items(self, response):
        response = response.json()
        entries = []

        for entry in response:
            update_timestamp = dateutil.parser.parse(entry['updateDate'])
            last_update_timestamp = dateutil.parser.parse(self.last_update['updateDate']) \
                if self.last_update else None

            if self.last_update is None \
                    or (update_timestamp > last_update_timestamp
                        and not (
                            entry['booker']['email'] == self.last_update['booker']['email']
                            and entry['status'] == self.last_update['status'])):
                self.last_update = entry
                entries.append(entry)

        return NewReservationEvent(dict(self), entries)


# vim:sw=4:ts=4:et:




from flask import Blueprint, request, render_template

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate
from platypush.backend.http.utils import HttpUtils
from platypush.config import Config
from platypush.user import UserManager

settings = Blueprint('settings', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    settings,
]


@settings.route('/settings', methods=['GET'])
@authenticate()
def settings():
    """ Settings page """
    user_manager = UserManager()
    users = user_manager.get_users()
    return render_template('settings/index.html', utils=HttpUtils, users=users, token=Config.get('token'))


# vim:sw=4:ts=4:et:
